from typing import Optional
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QFrame, QHBoxLayout, QStackedWidget, QWidget
from .side_bar import SideBar
from .side_bar_button import SideBarButton
from ...core import WidgetBase


class _SideBarWidget(QWidget):
    def __init__(self, 
                 parent: Optional[QWidget] = None,
                 stacked_widget: QStackedWidget = None) -> None:
        super().__init__(parent)

        self.setAttribute(Qt.WidgetAttribute.WA_StyledBackground, True)

        # Layout for this widget
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        # Background frame and its layout
        self._bg_frame = QFrame(self)

        self._bg_frame_layout = QHBoxLayout(self._bg_frame)
        self._bg_frame_layout.setContentsMargins(5, 5, 5, 5)
        self._bg_frame_layout.setSpacing(10)
        self._layout.addWidget(self._bg_frame)

        # Menu bar
        self._side_bar = SideBar(self)
        self._side_bar.setFixedWidth(60)
        self._bg_frame_layout.addWidget(self._side_bar)

        # Stacked widget and its layout
        if stacked_widget is not None:
            self._stacked_widget = stacked_widget
            self._stacked_widget.setParent(self)
        else:
            self._stacked_widget = QStackedWidget(self)

        self._bg_frame_layout.addWidget(self._stacked_widget)

    def addPage(self, 
                page: QWidget,
                button: SideBarButton, 
                button_alignment: SideBarButton.Alignment = SideBarButton.Alignment.top) -> None:

        if not hasattr(button, "clicked"):
            raise ValueError("invalid button value!")

        self._side_bar.addButton(button, button_alignment)
        self._stacked_widget.addWidget(page)
        button.clicked.connect(lambda: self._stacked_widget.setCurrentWidget(page))  # type: ignore[attr-defined]


class SideBarItemWrapper:
    def __init__(self,
                 *,
                 page: QWidget,
                 button: SideBarButton | QWidget,
                 alignment: SideBarButton.Alignment = SideBarButton.Alignment.top):
        self.page = page
        self.button = button
        self.alignment = alignment


class SideBarWidget(WidgetBase, _SideBarWidget):
    def __init__(self, 
                 *,
                 children: list[SideBarItemWrapper] = None,
                 stacked_widget: QStackedWidget = None,
                 **kwargs):
        _SideBarWidget.__init__(self, None, stacked_widget)
        super().__init__(**kwargs)

        if children is not None:
            for child in children:
                self.addPage(
                    child.page,
                    child.button,
                    child.alignment
                )


from typing import Tuple, Any
from PySide6.QtQuickWidgets import QQuickWidget
from PySide6.QtCore import QUrl, Qt
from ...core import WidgetBase


class QuickWidget(WidgetBase, QQuickWidget):
    def __init__(self, 
                 *,
                 qml: QUrl | str,
                 context: list[Tuple[str, Any]] = None,
                 **kwargs):
        QQuickWidget.__init__(self, qml)
        super().__init__(**kwargs)

        self._context = context

        self.setUpdatesEnabled(True)
        self.setTabletTracking(True)
        self.setMouseTracking(True)

        self.setResizeMode(QQuickWidget.ResizeMode.SizeRootObjectToView)
        self.setAttribute(Qt.WidgetAttribute.WA_X11NetWmWindowTypeDesktop, True)
        self.statusChanged.connect(self.__on_statusChanged)

    @property
    def root_object(self):
        return self.rootObject()

    @property
    def root_context(self):
        return self.rootContext()

    def __on_statusChanged(self, status):
        if self._context is not None:
            for (name, value) in self._context:
                self.rootContext().setContextProperty(name, value)

        if status == self.Status.Error:
            for error in self.errors():
                print("qml-widget-error: ", error.toString())


from typing import Tuple, Union
from PySide6.QtCore import QMargins
from PySide6.QtWidgets import QFrame, QLayout, QWidget
from ..core import WidgetBase


class Container(WidgetBase, QFrame):
    def __init__(self, 
                 *,
                 child: Union[QWidget, QLayout] = None,
                 padding: Union[QMargins, Tuple[int, int, int, int]] = None,
                 **kwargs):
        QFrame.__init__(self)
        super().__init__(**kwargs)

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setFrameShadow(QFrame.Shadow.Raised)

        if child is not None:
            if isinstance(child, QWidget):
                child.setParent(self)
                if padding is not None:
                    self.setContentsMargins(*padding
                                            if isinstance(padding, (tuple, list))
                                            else padding)

            elif isinstance(child, QLayout):
                child.setProperty("parent", self)
                self.setLayout(child)
                if padding is not None:
                    self.setContentsMargins(*padding 
                                            if isinstance(padding, (tuple, list)) 
                                            else padding)
            else:
                raise ValueError("the 'child' argument must be one of the QWidget or QLayout instance.")


class Frame(Container):
    pass


from PySide6.QtCore import Qt


class Modifiers:
    @staticmethod
    def is_shift(event):
        return event.modifiers() == Qt.KeyboardModifier.ShiftModifier

    @staticmethod
    def is_alt(event):
        return event.modifiers() == Qt.KeyboardModifier.AltModifier

    @staticmethod
    def is_ctrl(event):
        return event.modifiers() == Qt.KeyboardModifier.ControlModifier

    @staticmethod
    def is_meta(event):
        return event.modifiers() == Qt.KeyboardModifier.MetaModifier

    @staticmethod
    def is_win(event):
        return event.modifiers() == Qt.KeyboardModifier.MetaModifier



from typing import Union
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QLayoutItem, QSpacerItem, QVBoxLayout, QWidget, QLayout
from .item_wrapper import VLayoutItemWrapper
from ...core import QObjectBase


class VLayout(QObjectBase, QVBoxLayout):
    def __init__(self,
                 *,
                 children: list[Union[QWidget, QLayout, VLayoutItemWrapper]] = None,
                 **kwargs):
        QVBoxLayout.__init__(self)
        super().__init__(**kwargs)

        self._set_children(children)

    def _set_children(self, children: list[Union[QWidget, QLayout, 
                                                 VLayoutItemWrapper]]) -> None:
        if not children:
            return

        for child in children:
            if isinstance(child, QSpacerItem):
                self.addSpacerItem(child)

            elif isinstance(child, QLayoutItem):
                self.addItem(child)

            elif isinstance(child, QWidget):
                self.addWidget(child)

            elif isinstance(child, QLayout):
                self.addLayout(child)

            elif isinstance(child, VLayoutItemWrapper):
                if isinstance(child.child, QWidget):
                    self.addWidget(*child._yield_attr())

                elif isinstance(child.child, QLayout):
                    self.addLayout(*child._yield_attr())

    @staticmethod
    def add(
            child: QWidget | QLayout,
            stretch: int = None,
            alignment: Qt.AlignmentFlag = None
        ) -> VLayoutItemWrapper:
        return VLayoutItemWrapper(
            child,
            stretch,
            alignment
        )


from typing import Union
from PySide6.QtWidgets import QFormLayout, QWidget, QLayout, QLayoutItem
from .item_wrapper import FormLayoutItemWrapper
from ...core import QObjectBase


class FormLayout(QObjectBase, QFormLayout):
    def __init__(self,
                 *,
                 children: list[Union[QWidget, 
                                      QLayoutItem, 
                                      FormLayoutItemWrapper]],
                 **kwargs):
        QFormLayout.__init__(self)
        super().__init__(**kwargs)

        self._set_children(children)

    def _set_children(self, children: list[Union[QWidget, 
                                                 QLayoutItem, 
                                                 FormLayoutItemWrapper]]) -> None:
        if not children:
            return

        for child in children:
            if isinstance(child, QLayoutItem):
                self.addItem(child)

            elif isinstance(child, QWidget):
                self.addWidget(child)

            elif isinstance(child, FormLayoutItemWrapper):
                self.addRow(*child._yield_attr())

    @staticmethod
    def wrapper(
            label: Union[QWidget, str],
            field: Union[QWidget, QLayout]
        ) -> FormLayoutItemWrapper:
        return FormLayoutItemWrapper(
            label,
            field
        )

    # @staticmethod
    # def wrapper(
    #         row: Union[QWidget, QLayout]
    #     ) -> FormLayoutItemWrapper:
    #     return FormLayoutItemWrapper(row)


# -*- coding: utf-8 -*-

from PySide6.QtWidgets import QApplication, QWidget
from PySide6.QtCore import QRect, Qt, Signal
from PySide6.QtGui import (QClipboard, QPainter, QFont, 
                           QBrush, QColor, QPen, 
                           QContextMenuEvent)
from .backend import Session
from ...core import WidgetBase

DEBUG = False


class _TerminalWidget(QWidget):
    foreground_color_map = {
        0: "#000",
        1: "#b00",
        2: "#0b0",
        3: "#bb0",
        4: "#00b",
        5: "#b0b",
        6: "#0bb",
        7: "#bbb",
        8: "#666",
        9: "#f00",
        10: "#0f0",
        11: "#ff0",
        12: "#00f",  # concealed
        13: "#f0f",
        14: "#000",  # negative
        15: "#fff",  # default
    }
    background_color_map = {
        0: "#000",
        1: "#b00",
        2: "#0b0",
        3: "#bb0",
        4: "#00b",
        5: "#b0b",
        6: "#0bb",
        7: "#bbb",
        12: "#aaa",  # cursor
        14: "#000",  # default
        15: "#fff",  # negative
    }
    keymap = {
        Qt.Key_Backspace: chr(127),
        Qt.Key_Escape: chr(27),
        Qt.Key_AsciiTilde: chr(126),
        Qt.Key_Up: "~A",
        Qt.Key_Down: "~B",
        Qt.Key_Left: "~D",
        Qt.Key_Right: "~C",
        Qt.Key_PageUp: "~1",
        Qt.Key_PageDown: "~2",
        Qt.Key_Home: "~H",
        Qt.Key_End: "~F",
        Qt.Key_Insert: "~3",
        Qt.Key_Delete: "~4",
        Qt.Key_F1: "~a",
        Qt.Key_F2: "~b",
        Qt.Key_F3: "~c",
        Qt.Key_F4: "~d",
        Qt.Key_F5: "~e",
        Qt.Key_F6: "~f",
        Qt.Key_F7: "~g",
        Qt.Key_F8: "~h",
        Qt.Key_F9: "~i",
        Qt.Key_F10: "~j",
        Qt.Key_F11: "~k",
        Qt.Key_F12: "~l",
    }

    session_closed = Signal()

    def __init__(self, 
                 parent=None, 
                 command="/bin/bash",
                 font: QFont = QFont("Monospace", 18)):
        super(_TerminalWidget, self).__init__(parent)

        if self.parent() is not None:
            self.parent().setTabOrder(self, self)

        self.setAutoFillBackground(False)
        self.setCursor(Qt.CursorShape.IBeamCursor)
        self.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
        self.setAttribute(Qt.WidgetAttribute.WA_OpaquePaintEvent, True)
        self.setFont(font)

        self._session = None
        self._last_update = None
        self._screen = []
        self._text = []
        self._cursor_rect = None
        self._cursor_col = 0
        self._cursor_row = 0
        self._dirty = False
        self._blink = False
        self._press_pos = None
        self._selection = None
        self._clipboard = QApplication.clipboard()
        QApplication.instance().lastWindowClosed.connect(Session.close_all)

        if command:
            self.execute(command)

    def execute(self, command="/bin/bash"):
        self._session = Session()
        self._session.start(command)
        self._timer_id = None

        # start timer either with high or low priority
        if self.hasFocus():
            self.focusInEvent(None)
        else:
            self.focusOutEvent(None)

    def send(self, s):
        self._session.write(s)

    def stop(self):
        self._session.stop()

    def pid(self):
        return self._session.pid()

    def setFont(self, font):
        super(_TerminalWidget, self).setFont(font)
        self._update_metrics()

    def focusNextPrevChild(self, next):
        if not self._session.is_alive():
            return True
        return False

    def focusInEvent(self, event):
        if not self._session.is_alive():
            return

        if self._timer_id is not None:
            self.killTimer(self._timer_id)

        self._timer_id = self.startTimer(250)
        self.update_screen()

    def focusOutEvent(self, event):
        if not self._session.is_alive():
            return

        # reduced update interval 
        # -> slower screen updates
        # -> but less load on main app which results in better responsiveness
        if self._timer_id is not None:
            self.killTimer(self._timer_id)
        self._timer_id = self.startTimer(750)

    def resizeEvent(self, event):
        if not self._session.is_alive():
            return

        self._columns, self._rows = self._pixel2pos(self.width(), self.height())
        self._session.resize(self._columns, self._rows)
        self.update_screen()

    def closeEvent(self, event):
        if not self._session.is_alive():
            return
        self._session.close()

    def timerEvent(self, event):
        if not self._session.is_alive():
            if self._timer_id is not None:
                self.killTimer(self._timer_id)
                self._timer_id = None

            if DEBUG:
                print("Session closed")

            self.session_closed.emit()
            return

        last_change = self._session.last_change()
        if not last_change:
            return

        if not self._last_update or last_change > self._last_update:
            self._last_update = last_change
            old_screen = self._screen
            (self._cursor_col, self._cursor_row), self._screen = self._session.dump()
            self._update_cursor_rect()
            if old_screen != self._screen:
                self._dirty = True

        if self.hasFocus():
            self._blink = not self._blink

        self.update()

    def _update_metrics(self):
        fm = self.fontMetrics()
        self._char_height = fm.height()
        self._char_width = fm.averageCharWidth()

    def _update_cursor_rect(self):
        cx, cy = self._pos2pixel(self._cursor_col, self._cursor_row)
        self._cursor_rect = QRect(cx, cy, self._char_width, self._char_height)

    def _reset(self):
        self._update_metrics()
        self._update_cursor_rect()
        self.resizeEvent(None)
        self.update_screen()

    def update_screen(self):
        self._dirty = True
        self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        if self._dirty:
            self._dirty = False
            self._paint_screen(painter)
        else:
            if self._cursor_rect and not self._selection:
                self._paint_cursor(painter)
        if self._selection:
            self._paint_selection(painter)
            self._dirty = True

    def _pixel2pos(self, x, y):
        col = int(round(x / self._char_width))
        row = int(round(y / self._char_height))
        return col, row

    def _pos2pixel(self, col, row):
        x = col * self._char_width
        y = row * self._char_height
        return x, y

    def _paint_cursor(self, painter):
        if self._blink:
            color = "#aaa"
        else:
            color = "#fff"

        painter.setPen(QPen(QColor(color)))
        painter.drawRect(self._cursor_rect)

    def _paint_screen(self, painter):
        # Speed hacks: local name lookups are faster
        vars().update(QColor=QColor, QBrush=QBrush, QPen=QPen, QRect=QRect)

        background_color_map = self.background_color_map
        foreground_color_map = self.foreground_color_map
        char_width = self._char_width
        char_height = self._char_height
        painter_drawText = painter.drawText
        painter_fillRect = painter.fillRect
        painter_setPen = painter.setPen
        align = Qt.AlignTop | Qt.AlignLeft

        # set defaults
        background_color = background_color_map[14]
        foreground_color = foreground_color_map[15]
        brush = QBrush(QColor(background_color))
        painter_fillRect(self.rect(), brush)
        pen = QPen(QColor(foreground_color))
        painter_setPen(pen)
        y = 0
        text = []
        text_append = text.append
        for row, line in enumerate(self._screen):
            col = 0
            text_line = ""
            for item in line:
                if isinstance(item, str):
                    x = col * char_width
                    length = len(item)
                    rect = QRect(x, y, x + char_width * length, y + char_height)
                    painter_fillRect(rect, brush)
                    painter_drawText(rect, align, item)
                    col += length
                    text_line += item
                else:
                    foreground_color_idx, background_color_idx, underline_flag = item
                    foreground_color = foreground_color_map[foreground_color_idx]
                    background_color = background_color_map[background_color_idx]
                    pen = QPen(QColor(foreground_color))
                    brush = QBrush(QColor(background_color))
                    painter_setPen(pen)
                    # painter.setBrush(brush)
            y += char_height
            text_append(text_line)
        self._text = text

    def _paint_selection(self, painter):
        pcol = QColor(200, 200, 200, 50)
        pen = QPen(pcol)
        bcol = QColor(230, 230, 230, 50)
        brush = QBrush(bcol)
        painter.setPen(pen)
        painter.setBrush(brush)
        for (start_col, start_row, end_col, end_row) in self._selection:
            x, y = self._pos2pixel(start_col, start_row)
            width, height = self._pos2pixel(end_col - start_col, end_row - start_row)
            rect = QRect(x, y, width, height)
            # painter.drawRect(rect)
            painter.fillRect(rect, brush)

    def zoom_in(self):
        font = self.font()
        font.setPixelSize(font.pixelSize() + 2)
        self.setFont(font)
        self._reset()

    def zoom_out(self):
        font = self.font()
        font.setPixelSize(font.pixelSize() - 2)
        self.setFont(font)
        self._reset()

    return_pressed = Signal()

    def keyPressEvent(self, event):
        text = str(event.text())
        key = event.key()
        modifiers = event.modifiers()
        ctrl = modifiers == Qt.ControlModifier
        if ctrl and key == Qt.Key_Plus:
            self.zoom_in()
        elif ctrl and key == Qt.Key_Minus:
            self.zoom_out()
        else:
            if text and key != Qt.Key_Backspace:
                self.send(text.encode("utf-8"))
            else:
                s = self.keymap.get(key)
                if s:
                    self.send(s.encode("utf-8"))
                elif DEBUG:
                    print("Unknown key combination")
                    print("Modifiers:", modifiers)
                    print("Key:", key)
                    for name in dir(Qt):
                        if not name.startswith("Key_"):
                            continue
                        value = getattr(Qt, name)
                        if value == key:
                            print("Symbol: Qt.%s" % name)
                    print("Text: %r" % text)
        event.accept()
        if key in (Qt.Key_Enter, Qt.Key_Return):
            self.return_pressed.emit()

    def mousePressEvent(self, event):
        button = event.button()
        if button == Qt.RightButton:
            ctx_event = QContextMenuEvent(QContextMenuEvent.Mouse, event.pos())
            self.contextMenuEvent(ctx_event)
            self._press_pos = None

        elif button == Qt.LeftButton:
            self._press_pos = event.pos()
            self._selection = None
            self.update_screen()

        elif button == Qt.MiddleButton:
            self._press_pos = None
            self._selection = None
            text = str(self._clipboard.text(QClipboard.Selection))
            self.send(text.encode("utf-8"))
            # self.update_screen()

    def mouseReleaseEvent(self, QMouseEvent):
        if self._press_pos is not None:
            self._press_pos = None

        # pass  # self.update_screen()

    def _selection_rects(self, start_pos, end_pos):
        sx, sy = start_pos.x(), start_pos.y()
        start_col, start_row = self._pixel2pos(sx, sy)
        ex, ey = end_pos.x(), end_pos.y()
        end_col, end_row = self._pixel2pos(ex, ey)
        if start_row == end_row:
            if ey > sy or end_row == 0:
                end_row += 1
            else:
                end_row -= 1
        if start_col == end_col:
            if ex > sx or end_col == 0:
                end_col += 1
            else:
                end_col -= 1
        if start_row > end_row:
            start_row, end_row = end_row, start_row
        if start_col > end_col:
            start_col, end_col = end_col, start_col
        if end_row - start_row == 1:
            return [(start_col, start_row, end_col, end_row)]
        else:
            return [
                (start_col, start_row, self._columns, start_row + 1),
                (0, start_row + 1, self._columns, end_row - 1),
                (0, end_row - 1, end_col, end_row)
            ]

    def text(self, rect=None):
        if rect is None:
            return "\n".join(self._text)
        else:
            text = []
            (start_col, start_row, end_col, end_row) = rect
            for row in range(start_row, end_row):
                text.append(self._text[row][start_col:end_col])
            return text

    def text_selection(self):
        text = []
        for (start_col, start_row, end_col, end_row) in self._selection:
            for row in range(start_row, end_row):
                text.append(self._text[row][start_col:end_col])
        return "\n".join(text)

    def column_count(self):
        return self._columns

    def row_count(self):
        return self._rows

    def mouseMoveEvent(self, event):
        if self._press_pos:
            move_pos = event.pos()
            self._selection = self._selection_rects(self._press_pos, move_pos)

            sel = self.text_selection()
            if DEBUG:
                print("%r copied to xselection" % sel)
            self._clipboard.setText(sel, QClipboard.Selection)

            self.update_screen()    

    def mouseDoubleClickEvent(self, event):
        self._press_pos = None
        # double clicks create a selection for the word under the cursor
        pos = event.pos()
        x, y = pos.x(), pos.y()
        col, row = self._pixel2pos(x, y)
        line = self._text[row]
        # find start of word
        start_col = col
        found_left = 0
        while start_col > 0:
            char = line[start_col]
            if not char.isalnum() and char not in ("_",):
                found_left = 1
                break
            start_col -= 1
        # find end of word
        end_col = col
        found_right = 0
        while end_col < self._columns:
            char = line[end_col]
            if not char.isalnum() and char not in ("_",):
                found_right = 1
                break
            end_col += 1
        self._selection = [(start_col + found_left, 
                            row, 
                            end_col - found_right + 1, 
                            row + 1)]

        sel = self.text_selection()
        if DEBUG:
            print("%r copied to xselection" % sel)
        self._clipboard.setText(sel, QClipboard.Selection)

        self.update_screen()

    def is_alive(self):
        return (self._session and self._session.is_alive()) or False


class TerminalWidget(WidgetBase, _TerminalWidget):
    def __init__(self,
                 *,
                 cmd: str = "/bin/bash",
                 font: QFont = QFont("Monospace", 12),
                 **kwargs):
        _TerminalWidget.__init__(self, command=cmd, font=font)
        super().__init__(**kwargs)


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget
from ..core.base import BehaviorDeclarative


class Alignment(BehaviorDeclarative):
    def __init__(self,
                 *,
                 child: QWidget,
                 alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter):

        if not hasattr(child, "setAlignment"):
            raise ValueError("invalid child widget!")

        child.setAlignment(alignment)
        return child


from winreg import HKEY_CURRENT_USER as hkey, QueryValueEx as getSubkeyValue, OpenKey as getKey
from ...enums.theme import Theme

import ctypes
import ctypes.wintypes

advapi32 = ctypes.windll.advapi32

# LSTATUS RegOpenKeyExA(
#     HKEY hKey,
#     LPCSTR lpSubKey,
#     DWORD ulOptions,
#     REGSAM samDesired,
#     PHKEY phkResult
# );
advapi32.RegOpenKeyExA.argtypes = (
    ctypes.wintypes.HKEY,
    ctypes.wintypes.LPCSTR,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.DWORD,
    ctypes.POINTER(ctypes.wintypes.HKEY),
)
advapi32.RegOpenKeyExA.restype = ctypes.wintypes.LONG

# LSTATUS RegQueryValueExA(
#     HKEY hKey,
#     LPCSTR lpValueName,
#     LPDWORD lpReserved,
#     LPDWORD lpType,
#     LPBYTE lpData,
#     LPDWORD lpcbData
# );
advapi32.RegQueryValueExA.argtypes = (
    ctypes.wintypes.HKEY,
    ctypes.wintypes.LPCSTR,
    ctypes.wintypes.LPDWORD,
    ctypes.wintypes.LPDWORD,
    ctypes.wintypes.LPBYTE,
    ctypes.wintypes.LPDWORD,
)
advapi32.RegQueryValueExA.restype = ctypes.wintypes.LONG

# LSTATUS RegNotifyChangeKeyValue(
#     HKEY hKey,
#     WINBOOL bWatchSubtree,
#     DWORD dwNotifyFilter,
#     HANDLE hEvent,
#     WINBOOL fAsynchronous
# );
advapi32.RegNotifyChangeKeyValue.argtypes = (
    ctypes.wintypes.HKEY,
    ctypes.wintypes.BOOL,
    ctypes.wintypes.DWORD,
    ctypes.wintypes.HANDLE,
    ctypes.wintypes.BOOL,
)
advapi32.RegNotifyChangeKeyValue.restype = ctypes.wintypes.LONG

def theme() -> Theme:
    """ Uses the Windows Registry to detect if the user is using Dark Mode """
    # Registry will return 0 if Windows is in Dark Mode and 1 if Windows is in Light Mode. This dictionary converts that output into the text that the program is expecting.
    valueMeaning = {0: Theme.dark, 1: Theme.light}
    # In HKEY_CURRENT_USER, get the Personalisation Key.
    try:
        key = getKey(hkey, "Software\\Microsoft\\Windows\\CurrentVersion\\Themes\\Personalize")
        # In the Personalisation Key, get the AppsUseLightTheme subkey. This returns a tuple.
        # The first item in the tuple is the result we want (0 or 1 indicating Dark Mode or Light Mode); the other value is the type of subkey e.g. DWORD, QWORD, String, etc.
        subkey = getSubkeyValue(key, "AppsUseLightTheme")[0]
    except FileNotFoundError:
        # some headless Windows instances (e.g. GitHub Actions or Docker images) do not have this key
        return None
    return valueMeaning[subkey]

def isDark():
    if theme() is not None:
        return theme() == Theme.dark

def isLight():
    if theme() is not None:
        return theme() == Theme.light

#def listener(callback: typing.Callable[[enums.Theme], None]) -> None:
def listener(callback):
    hKey = ctypes.wintypes.HKEY()
    advapi32.RegOpenKeyExA(
        ctypes.wintypes.HKEY(0x80000001), # HKEY_CURRENT_USER
        ctypes.wintypes.LPCSTR(b'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Themes\\Personalize'),
        ctypes.wintypes.DWORD(),
        ctypes.wintypes.DWORD(0x00020019), # KEY_READ
        ctypes.byref(hKey),
    )

    dwSize = ctypes.wintypes.DWORD(ctypes.sizeof(ctypes.wintypes.DWORD))
    queryValueLast = ctypes.wintypes.DWORD()
    queryValue = ctypes.wintypes.DWORD()
    advapi32.RegQueryValueExA(
        hKey,
        ctypes.wintypes.LPCSTR(b'AppsUseLightTheme'),
        ctypes.wintypes.LPDWORD(),
        ctypes.wintypes.LPDWORD(),
        ctypes.cast(ctypes.byref(queryValueLast), ctypes.wintypes.LPBYTE),
        ctypes.byref(dwSize),
    )

    while True:
        advapi32.RegNotifyChangeKeyValue(
            hKey,
            ctypes.wintypes.BOOL(True),
            ctypes.wintypes.DWORD(0x00000004), # REG_NOTIFY_CHANGE_LAST_SET
            ctypes.wintypes.HANDLE(None),
            ctypes.wintypes.BOOL(False),
        )
        advapi32.RegQueryValueExA(
            hKey,
            ctypes.wintypes.LPCSTR(b'AppsUseLightTheme'),
            ctypes.wintypes.LPDWORD(),
            ctypes.wintypes.LPDWORD(),
            ctypes.cast(ctypes.byref(queryValue), ctypes.wintypes.LPBYTE),
            ctypes.byref(dwSize),
        )
        if queryValueLast.value != queryValue.value:
            queryValueLast.value = queryValue.value
            callback(Theme.light if queryValue.value else Theme.dark)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import math

from PySide6.QtCore import QPoint, QRect, QSizeF, QTimer, Qt, QRectF, QSize
from PySide6.QtGui import QPainter, QPainterPath, QColor
from PySide6.QtWidgets import QProgressBar
from ..core import WidgetBase


class _WaterRippleProgressBar(QProgressBar):
    water_height: int = 1
    water_density: int = 1
    style_type: int = 1
    border_radius: int = 8
    text_color: Qt.GlobalColor = Qt.GlobalColor.white
    bg_color: Qt.GlobalColor = Qt.GlobalColor.gray
    fg_water_color: QColor = QColor(33, 178, 148)
    bg_water_color: QColor = QColor(33, 178, 148, 100)

    def __init__(self) -> None:
        super(_WaterRippleProgressBar, self).__init__()

        self._offset = 0

        self._updateTimer = QTimer(self, timeout=self.update)
        self._updateTimer.start(100)

    def setRange(self, min: int, max: int):
        if min == max == 0:
            return

        super(_WaterRippleProgressBar, self).setRange(min, max)

    def setMinimum(self, value: int):
        if value == self.maximum() == 0:
            return

        super(_WaterRippleProgressBar, self).setMinimum(value)

    def setMaximum(self, value):
        if value == self.minimum() == 0:
            return

        super(_WaterRippleProgressBar, self).setMaximum(value)

    def setWaterHeight(self, height):
        self.water_height = height
        self.update()

    def setWaterDensity(self, density):
        self.water_density = density
        self.update()

    def setStyleType(self, style):
        self.style_type = style
        self.update()

    def setBorderRadius(self, border_radius: int):
        self.border_radius = border_radius
        self.update()

    def sizeHint(self):
        return QSize(100, 100)

    def paintEvent(self, event):
        if self.minimum() == self.maximum() == 0:
            return

        percent = 1 - (self.value() - self.minimum()) / \
                  (self.maximum() - self.minimum())

        w = 6 * self.water_density * math.pi / self.width()
        A = self.height() * self.water_height * 1 / 26
        k = self.height() * percent

        waterPath1 = QPainterPath()
        waterPath1.moveTo(0, self.height())
        waterPath2 = QPainterPath()
        waterPath2.moveTo(0, self.height())

        self._offset += 0.6
        if self._offset > self.width() / 2:
            self._offset = 0

        for i in range(self.width() + 1):
            y = A * math.sin(w * i + self._offset) + k
            waterPath1.lineTo(i, y)

            y = A * math.sin(w * i + self._offset + self.width() / 2 * A) + k
            waterPath2.lineTo(i, y)

        waterPath1.lineTo(self.width(), self.height())
        waterPath1.lineTo(0, self.height())
        waterPath2.lineTo(self.width(), self.height())
        waterPath2.lineTo(0, self.height())

        bgPath = QPainterPath()
        if self.style_type and self.border_radius:
            bgPath.addRoundedRect(QRectF(self.rect()), 
                                  self.border_radius,
                                  self.border_radius)
        elif self.style_type:
            bgPath.addRect(QRectF(self.rect()))
        else:
            radius = min(self.width(), self.height())
            bgPath.addRoundedRect(QRectF(self.rect()), radius, radius)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setPen(Qt.PenStyle.NoPen)

        if (not self.style_type
            or (self.style_type and self.border_radius)):
            painter.setClipPath(bgPath)

        painter.save()
        painter.setBrush(self.bg_color)
        painter.drawPath(bgPath)
        painter.restore()

        painter.save()
        painter.setBrush(self.bg_water_color)
        painter.drawPath(waterPath1)
        painter.restore()

        painter.save()
        painter.setBrush(self.fg_water_color)
        painter.drawPath(waterPath2)
        painter.restore()

        if self.isTextVisible():
            pixelRatio = self.devicePixelRatioF()
            sz = QSizeF(self.width() * pixelRatio, self.height() * pixelRatio).toSize()
            rect = QRectF(0, 0, self.width() * pixelRatio, self.height() * pixelRatio)

            font = painter.font()
            rectValue = QRect()
            progressText = self.text().strip('%')

            if progressText == '100':
                font.setPixelSize(sz.height() * 35 / 100)
                painter.setFont(font)

                rectValue.setWidth(sz.width() * 60 / 100)
                rectValue.setHeight(sz.height() * 35 / 100)
                rectValue.moveCenter(rect.center().toPoint())

                painter.setPen(self.text_color)
                painter.drawText(rectValue, Qt.AlignmentFlag.AlignCenter, progressText)
            else:
                font.setPixelSize(sz.height() * 40 / 100)
                painter.setFont(font)

                rectValue.setWidth(sz.width() * 45 / 100)
                rectValue.setHeight(sz.height() * 40 / 100)
                rectValue.moveCenter(rect.center().toPoint())
                rectValue.moveLeft(rect.left() + rect.width() * 0.45 * 0.5)

                painter.setPen(self.text_color)
                painter.drawText(rectValue, Qt.AlignmentFlag.AlignCenter, progressText)

                font.setPixelSize(font.pixelSize() / 2)
                painter.setFont(font)
                rectPerent = QRect(QPoint(rectValue.right(), rectValue.bottom() - rect.height() * 20 / 100),
                                   QPoint(rectValue.right() + rect.width() * 20 / 100, rectValue.bottom()))

                painter.drawText(rectPerent, Qt.AlignmentFlag.AlignCenter, '%')


class WaterRippleProgressBar(WidgetBase, _WaterRippleProgressBar):
    '''
    :param:style_type: 0 = rounded, 1 = rectangle
    '''
    def __init__(self,
                 *,
                 water_height: int = 1,
                 water_density: int = 1,
                 style_type: int = 1,
                 border_radius: int = 8,
                 text_color: Qt.GlobalColor = Qt.GlobalColor.white,
                 bg_color: Qt.GlobalColor = Qt.GlobalColor.gray,
                 fg_water_color: QColor = QColor(33, 178, 148),
                 bg_water_color: QColor = QColor(33, 178, 148, 100),
                 **kwargs):
        _WaterRippleProgressBar.__init__(self)
        super().__init__(**kwargs)

        self.water_height = water_height
        self.water_density = water_density
        self.style_type = style_type
        self.border_radius = border_radius
        self.text_color = text_color
        self.bg_color = bg_color
        self.fg_water_color = fg_water_color
        self.bg_water_color = bg_water_color


from PySide6.QtCore import QFile


class File(QFile):
    def __init__(self, 
                 file: str,
                 mode: QFile.OpenModeFlag = QFile.OpenModeFlag.ReadOnly):
        super().__init__(file)
        self._mode = mode

    def __enter__(self):
        self.open(self._mode)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback) -> None:
        self.close()
    
    def read_with_utf_8(self) -> str:
        return str(self.readAll(), 
                   encoding='utf-8')


from typing import Union
from PySide6.QtCore import QMimeData
from PySide6.QtWidgets import QApplication
from PySide6.QtGui import QClipboard, QImage, QPixmap
from ..enums.clipboard import ClipboardDataTypes


class Clipboard:
    @staticmethod
    def paste_data(
        data_type: ClipboardDataTypes = ClipboardDataTypes.text,
        mode: QClipboard.Mode = QClipboard.Mode.Clipboard,
    ) -> Union[QMimeData, str, QImage, QPixmap]:

        cb = QApplication.clipboard()
        match data_type:
            case ClipboardDataTypes.image:
                return cb.image(mode)
            case ClipboardDataTypes.pixmap:
                return cb.pixmap(mode)
            case ClipboardDataTypes.mimedata:
                return cb.mimeData(mode)
            case ClipboardDataTypes.text:
                return cb.text(mode)

    @staticmethod
    def copy_data(
        data: Union[QMimeData, str, QImage, QPixmap],
        data_type: ClipboardDataTypes = ClipboardDataTypes.text,
        mode: QClipboard.Mode = QClipboard.Mode.Clipboard,
    ) -> None:

        cb = QApplication.clipboard()
        match data_type:
            case ClipboardDataTypes.image:
                return cb.setImage(data, mode)
            case ClipboardDataTypes.pixmap:
                return cb.setPixmap(data, mode)
            case ClipboardDataTypes.mimedata:
                return cb.setMimeData(data, mode)
            case ClipboardDataTypes.text:
                return cb.setText(data, mode)

    @staticmethod
    def copy(
        data: Union[QMimeData, str, QImage, QPixmap],
        data_type: ClipboardDataTypes = ClipboardDataTypes.text,
        mode: QClipboard.Mode = QClipboard.Mode.Clipboard,
    ) -> None:
        return Clipboard.copy_data(data, data_type, mode)

    @staticmethod
    def paste(
        data_type: ClipboardDataTypes = ClipboardDataTypes.text,
        mode: QClipboard.Mode = QClipboard.Mode.Clipboard,
    ) -> Union[QMimeData, str, QImage, QPixmap]:
        return Clipboard.paste_data(data_type, mode)


from typing import Union, List, Tuple, Callable, Any, TypeAlias 
from PySide6 import QtCore
from enum import Enum, auto


class Signals(Enum):
    about_to_activate: QtCore.Signal = auto()
    about_to_hide: QtCore.Signal = auto()
    about_to_quit: QtCore.Signal = auto()
    about_to_show: QtCore.Signal = auto()
    accepted: QtCore.Signal = auto()
    action_triggered: QtCore.Signal = auto()
    activated: QtCore.Signal = auto()
    allowed_areas_changed: QtCore.Signal = auto()
    anchor_clicked: QtCore.Signal = auto()
    angle_changed: QtCore.Signal = auto()
    application_display_name_changed: QtCore.Signal = auto()
    application_name_changed: QtCore.Signal = auto()
    application_state_changed: QtCore.Signal = auto()
    application_version_changed: QtCore.Signal = auto()
    axis_changed: QtCore.Signal = auto()
    backward_available: QtCore.Signal = auto()
    block_count_changed: QtCore.Signal = auto()
    blur_hints_changed: QtCore.Signal = auto()
    blur_radius_changed: QtCore.Signal = auto()
    button_clicked: QtCore.Signal = auto()
    button_pressed: QtCore.Signal = auto()
    button_released: QtCore.Signal = auto()
    button_toggled: QtCore.Signal = auto()
    canceled: QtCore.Signal = auto()
    cell_activated: QtCore.Signal = auto()
    cell_changed: QtCore.Signal = auto()
    cell_clicked: QtCore.Signal = auto()
    cell_double_clicked: QtCore.Signal = auto()
    cell_entered: QtCore.Signal = auto()
    cell_pressed: QtCore.Signal = auto()
    changed: QtCore.Signal = auto()
    checkable_changed: QtCore.Signal = auto()
    children_changed: QtCore.Signal = auto()
    clicked: QtCore.Signal = auto()
    close_editor: QtCore.Signal = auto()
    collapsed: QtCore.Signal = auto()
    color_changed: QtCore.Signal = auto()
    color_selected: QtCore.Signal = auto()
    columns_about_to_be_inserted: QtCore.Signal = auto()
    columns_about_to_be_moved: QtCore.Signal = auto()
    columns_about_to_be_removed: QtCore.Signal = auto()
    columns_inserted: QtCore.Signal = auto()
    columns_moved: QtCore.Signal = auto()
    columns_removed: QtCore.Signal = auto()
    commit_data: QtCore.Signal = auto()
    commit_data_request: QtCore.Signal = auto()
    complete_changed: QtCore.Signal = auto()
    copy_available: QtCore.Signal = auto()
    current_cell_changed: QtCore.Signal = auto()
    current_changed: QtCore.Signal = auto()
    current_char_format_changed: QtCore.Signal = auto()
    current_color_changed: QtCore.Signal = auto()
    current_font_changed: QtCore.Signal = auto()
    current_id_changed: QtCore.Signal = auto()
    current_index_changed: QtCore.Signal = auto()
    current_item_changed: QtCore.Signal = auto()
    current_page_changed: QtCore.Signal = auto()
    current_row_changed: QtCore.Signal = auto()
    current_text_changed: QtCore.Signal = auto()
    current_url_changed: QtCore.Signal = auto()
    cursor_position_changed: QtCore.Signal = auto()
    custom_button_clicked: QtCore.Signal = auto()
    custom_context_menu_requested: QtCore.Signal = auto()
    data_changed: QtCore.Signal = auto()
    date_changed: QtCore.Signal = auto()
    date_time_changed: QtCore.Signal = auto()
    destroyed: QtCore.Signal = auto()
    directory_entered: QtCore.Signal = auto()
    directory_loaded: QtCore.Signal = auto()
    directory_url_entered: QtCore.Signal = auto()
    dock_location_changed: QtCore.Signal = auto()
    document_size_changed: QtCore.Signal = auto()
    double_clicked: QtCore.Signal = auto()
    double_value_changed: QtCore.Signal = auto()
    double_value_selected: QtCore.Signal = auto()
    edit_text_changed: QtCore.Signal = auto()
    editing_finished: QtCore.Signal = auto()
    enabled_changed: QtCore.Signal = auto()
    entered: QtCore.Signal = auto()
    expanded: QtCore.Signal = auto()
    features_changed: QtCore.Signal = auto()
    file_renamed: QtCore.Signal = auto()
    file_selected: QtCore.Signal = auto()
    files_selected: QtCore.Signal = auto()
    filter_selected: QtCore.Signal = auto()
    finished: QtCore.Signal = auto()
    focus_changed: QtCore.Signal = auto()
    focus_item_changed: QtCore.Signal = auto()
    focus_object_changed: QtCore.Signal = auto()
    focus_window_changed: QtCore.Signal = auto()
    font_changed: QtCore.Signal = auto()
    font_database_changed: QtCore.Signal = auto()
    font_selected: QtCore.Signal = auto()
    forward_available: QtCore.Signal = auto()
    geometries_changed: QtCore.Signal = auto()
    geometry_changed: QtCore.Signal = auto()
    header_data_changed: QtCore.Signal = auto()
    height_changed: QtCore.Signal = auto()
    help_requested: QtCore.Signal = auto()
    highlighted: QtCore.Signal = auto()
    history_changed: QtCore.Signal = auto()
    hovered: QtCore.Signal = auto()
    icon_size_changed: QtCore.Signal = auto()
    id_clicked: QtCore.Signal = auto()
    id_pressed: QtCore.Signal = auto()
    id_released: QtCore.Signal = auto()
    id_toggled: QtCore.Signal = auto()
    indexes_moved: QtCore.Signal = auto()
    input_rejected: QtCore.Signal = auto()
    int_value_changed: QtCore.Signal = auto()
    int_value_selected: QtCore.Signal = auto()
    item_activated: QtCore.Signal = auto()
    item_changed: QtCore.Signal = auto()
    item_clicked: QtCore.Signal = auto()
    item_collapsed: QtCore.Signal = auto()
    item_double_clicked: QtCore.Signal = auto()
    item_entered: QtCore.Signal = auto()
    item_expanded: QtCore.Signal = auto()
    item_pressed: QtCore.Signal = auto()
    item_selection_changed: QtCore.Signal = auto()
    key_sequence_changed: QtCore.Signal = auto()
    last_window_closed: QtCore.Signal = auto()
    layout_about_to_be_changed: QtCore.Signal = auto()
    layout_changed: QtCore.Signal = auto()
    layout_direction_changed: QtCore.Signal = auto()
    link_activated: QtCore.Signal = auto()
    link_hovered: QtCore.Signal = auto()
    message_changed: QtCore.Signal = auto()
    message_clicked: QtCore.Signal = auto()
    model_about_to_be_reset: QtCore.Signal = auto()
    model_reset: QtCore.Signal = auto()
    modification_changed: QtCore.Signal = auto()
    movable_changed: QtCore.Signal = auto()
    object_name_changed: QtCore.Signal = auto()
    offset_changed: QtCore.Signal = auto()
    opacity_changed: QtCore.Signal = auto()
    opacity_mask_changed: QtCore.Signal = auto()
    organization_domain_changed: QtCore.Signal = auto()
    organization_name_changed: QtCore.Signal = auto()
    orientation_changed: QtCore.Signal = auto()
    origin_changed: QtCore.Signal = auto()
    overflow: QtCore.Signal = auto()
    page_added: QtCore.Signal = auto()
    page_count_changed: QtCore.Signal = auto()
    page_removed: QtCore.Signal = auto()
    palette_changed: QtCore.Signal = auto()
    parent_changed: QtCore.Signal = auto()
    pressed: QtCore.Signal = auto()
    primary_screen_changed: QtCore.Signal = auto()
    range_changed: QtCore.Signal = auto()
    redo_available: QtCore.Signal = auto()
    rejected: QtCore.Signal = auto()
    released: QtCore.Signal = auto()
    return_pressed: QtCore.Signal = auto()
    root_path_changed: QtCore.Signal = auto()
    rotation_changed: QtCore.Signal = auto()
    rows_about_to_be_inserted: QtCore.Signal = auto()
    rows_about_to_be_moved: QtCore.Signal = auto()
    rows_about_to_be_removed: QtCore.Signal = auto()
    rows_inserted: QtCore.Signal = auto()
    rows_moved: QtCore.Signal = auto()
    rows_removed: QtCore.Signal = auto()
    rubber_band_changed: QtCore.Signal = auto()
    save_state_request: QtCore.Signal = auto()
    scale_changed: QtCore.Signal = auto()
    scene_rect_changed: QtCore.Signal = auto()
    screen_added: QtCore.Signal = auto()
    screen_removed: QtCore.Signal = auto()
    scroller_properties_changed: QtCore.Signal = auto()
    section_clicked: QtCore.Signal = auto()
    section_count_changed: QtCore.Signal = auto()
    section_double_clicked: QtCore.Signal = auto()
    section_entered: QtCore.Signal = auto()
    section_handle_double_clicked: QtCore.Signal = auto()
    section_moved: QtCore.Signal = auto()
    section_pressed: QtCore.Signal = auto()
    section_resized: QtCore.Signal = auto()
    selection_changed: QtCore.Signal = auto()
    size_hint_changed: QtCore.Signal = auto()
    slider_moved: QtCore.Signal = auto()
    slider_pressed: QtCore.Signal = auto()
    slider_released: QtCore.Signal = auto()
    sort_indicator_changed: QtCore.Signal = auto()
    sort_indicator_clearable_changed: QtCore.Signal = auto()
    source_changed: QtCore.Signal = auto()
    splitter_moved: QtCore.Signal = auto()
    state_changed: QtCore.Signal = auto()
    strength_changed: QtCore.Signal = auto()
    sub_window_activated: QtCore.Signal = auto()
    tab_bar_clicked: QtCore.Signal = auto()
    tab_bar_double_clicked: QtCore.Signal = auto()
    tab_close_requested: QtCore.Signal = auto()
    tab_moved: QtCore.Signal = auto()
    tabified_dock_widget_activated: QtCore.Signal = auto()
    text_activated: QtCore.Signal = auto()
    text_changed: QtCore.Signal = auto()
    text_edited: QtCore.Signal = auto()
    text_highlighted: QtCore.Signal = auto()
    text_value_changed: QtCore.Signal = auto()
    text_value_selected: QtCore.Signal = auto()
    time_changed: QtCore.Signal = auto()
    toggled: QtCore.Signal = auto()
    tool_button_style_changed: QtCore.Signal = auto()
    top_level_changed: QtCore.Signal = auto()
    triggered: QtCore.Signal = auto()
    undo_available: QtCore.Signal = auto()
    update: QtCore.Signal = auto()
    update_block: QtCore.Signal = auto()
    update_preview_widget: QtCore.Signal = auto()
    update_request: QtCore.Signal = auto()
    url_selected: QtCore.Signal = auto()
    urls_selected: QtCore.Signal = auto()
    user_date_changed: QtCore.Signal = auto()
    user_time_changed: QtCore.Signal = auto()
    value_changed: QtCore.Signal = auto()
    viewport_entered: QtCore.Signal = auto()
    visibility_changed: QtCore.Signal = auto()
    visible_changed: QtCore.Signal = auto()
    widget_removed: QtCore.Signal = auto()
    width_changed: QtCore.Signal = auto()
    window_icon_changed: QtCore.Signal = auto()
    window_icon_text_changed: QtCore.Signal = auto()
    window_state_changed: QtCore.Signal = auto()
    window_title_changed: QtCore.Signal = auto()
    x_changed: QtCore.Signal = auto()
    x_scale_changed: QtCore.Signal = auto()
    y_changed: QtCore.Signal = auto()
    y_scale_changed: QtCore.Signal = auto()
    z_changed: QtCore.Signal = auto()
    z_scale_changed: QtCore.Signal = auto()

SignalTypeVar: TypeAlias = List[Tuple[Union[Signals, str], Callable[..., Any]]]



# coding: utf-8
class TextWrap:
    """ Text wrap """

    char_widths = [
        (126, 1), (159, 0), (687, 1), (710, 0),
        (711, 1), (727, 0), (733, 1), (879, 0),
        (1154, 1), (1161, 0), (4347, 1), (4447, 2),
        (7467, 1), (7521, 0), (8369, 1), (8426, 0),
        (9000, 1), (9002, 2), (11021, 1), (12350, 2),
        (12351, 1), (12438, 2), (12442, 0), (19893, 2),
        (19967, 1), (55203, 2), (63743, 1), (64106, 2),
        (65039, 1), (65059, 0), (65131, 2), (65279, 1),
        (65376, 2), (65500, 1), (65510, 2), (120831, 1),
        (262141, 2), (1114109, 1),
    ]

    @classmethod
    def get_width(cls, char):
        """Return the screen column width for a char"""
        o = ord(char)
        if o == 0xe or o == 0xf:
            return 0

        for num, wid in cls.char_widths:
            if o <= num:
                return wid

        return 1

    @classmethod
    def wrap(cls, text: str, width: int, once=True):
        """ Wrap according to string length

        Parameters
        ----------
        text: str
            the text to be wrapped

        width: int
            the maximum length of a single line, the length of Chinese characters is 2

        once: bool
            whether to wrap only once

        Returns
        -------
        wrap_text: str
            text after auto word wrap process

        is_wrapped: bool
            whether a line break occurs in the text
        """
        texts = text.strip().split('\n')
        result = []
        is_wrapped = False

        for text in texts:
            text_wrapped, wrapped = cls._wrap_line(text, width, once)
            is_wrapped |= wrapped
            result.append(text_wrapped)
            if once:
                result.extend(texts[1:])
                break

        return '\n'.join(result), is_wrapped

    @classmethod
    def _wrap_line(cls, text: str, width: int, once=True):
        count = 0
        last_count = 0
        chars = []
        is_wrapped = False
        break_pos = 0
        is_break_alpha = True
        n_inside_break = 0

        i = 0
        while i < len(text):
            c = text[i]
            length = cls.get_width(c)
            count += length

            # record the position of blank character
            if c == " " or length > 1:
                break_pos = i + n_inside_break
                last_count = count
                is_break_alpha = length == 1

            # No line breaks
            if count <= width:
                chars.append(c)
                i += 1
                continue

            # wrap at the position of the previous space
            if break_pos > 0 and is_break_alpha:
                if c != " ":
                    chars[break_pos] = '\n'
                    chars.append(c)

                    # Wrap inside long words
                    if last_count != 0:
                        count -= last_count
                        last_count = 0
                    else:
                        chars.insert(i, '\n')
                        break_pos = i
                        n_inside_break += 1
                else:
                    chars.append('\n')
                    count = 0
                    last_count = 0
            else:
                chars.extend(('\n', c))
                count = length

            is_wrapped = True

            # early return
            if once:
                return ''.join(chars)+text[i+1:], True

            i += 1

        return ''.join(chars), is_wrapped



from .widget import Widget
from .push_button import PushButton
from .wiggly_widget import WigglyWidget
from .eliding_label import ElidingLabel
from .qtermwidget import TerminalWidget
from .qml import QuickView, QuickWidget
from .application import Application, App

from .window import (
    MainWindow, 
    RoutingWindow, 
    FramelessWindow,
    FramelessWindowSizeGrip
)

from .side_bar import (
    SideBarItemWrapper,
    SideBarWidget,
    SideBarButton,
    SideBar
)

from .nav_bar import (
    NavBarItemWrapper,
    NavBarWidget,
    NavBarButton,
    NavBar
)

from .stacked_widget import (
    SlidingStackedWidget,
    StackedWidget
)

from .dialogs import (
    MaskDialog,
    SilentTextDialog,
    LargTextDialog
)

from .teaching_tip import TeachingTip
from .graphics_view import GraphicsView
from .container import Container, Frame
from .water_progress_bar import WaterProgressBar
from .water_ripple_progress_bar import WaterRippleProgressBar
from .colourful_progress import ColourfulProgressBar
from .metro_circle_progress_bar import MetroCircleProgress
from .progress_indicator import ProgressIndicator
from .label import Label
from .movie_view import MovieView
from .theme_switch_button import ThemeSwitchButton
from .line_edit import LineEdit
from .scroll_area import ScrollArea
from .icon_widget import IconWidget
from .lines import HLine, VLine
from .tool_button import ToolButton
from .menu import Menu, MenuSeparator, MenuSection, MenuSimpleAction


import unicodedata
from PySide6.QtCore import Qt


def slugify(original: str) -> str:
    """
    Make a string url friendly. Useful for creating routes for navigation.

    >>> slugify("What's    up?")
    'whats-up'

    >>> slugify("  Mitä kuuluu?  ")
    'mitä-kuuluu'
    """
    slugified = original.strip()
    slugified = " ".join(slugified.split())  # Remove extra spaces between words
    slugified = slugified.lower()
    # Remove unicode punctuation
    slugified = "".join(
        character
        for character in slugified
        if not unicodedata.category(character).startswith("P")
    )
    slugified = slugified.replace(" ", "-")

    return slugified


def qt_corner_to_edge(corner: Qt.Corner,
                      default: Qt.Edge = Qt.Edge(0)) -> Qt.Edge:
    return {
        Qt.Corner.TopLeftCorner: Qt.Edge.LeftEdge | Qt.Edge.TopEdge,
        Qt.Corner.TopRightCorner: Qt.Edge.RightEdge | Qt.Edge.TopEdge,
        Qt.Corner.BottomLeftCorner: Qt.Edge.LeftEdge | Qt.Edge.BottomEdge,
        Qt.Corner.BottomRightCorner: Qt.Edge.RightEdge | Qt.Edge.BottomEdge
    }.get(corner, default)


from .contrast import Contrast, ContrastFont
from .images import ImageColors
from .tools import (
    get_color_from_hex,
    get_hex_from_color,
    get_random_color,
    get_image_average_color,
    is_color_transparent,
    mixColor,
    mixDark,
    mixLight,
    print_hex_color,
    translucent
)


from typing import Union
from PySide6.QtCore import QPoint, QRect, QSize, Qt, Signal
from PySide6.QtGui import QCloseEvent, QMouseEvent, QResizeEvent, QScreen, QShowEvent
from PySide6.QtWidgets import (
    QApplication,
    QDockWidget, 
    QLayout, 
    QSizeGrip,
    QStatusBar,
    QSystemTrayIcon, 
    QVBoxLayout,
    QWidget, 
    QMainWindow,
    QSizePolicy
)

from ...core import WidgetBase
from ...utils.methods import qt_corner_to_edge


class FramelessWindowSizeGrip(QSizeGrip):
    def __init__(self,
                 *,
                 size: Union[tuple[int, int], QSize] = QSize(12, 12),
                 corner: Qt.Corner = Qt.Corner.BottomRightCorner) -> None:
        super().__init__(None)

        self._corner = corner
        self.resize(*size if not isinstance(size, QSize) else size)
        self.get_cursor_for_position()

    def get_cursor_for_position(self):
        self.setCursor({
            Qt.Corner.TopLeftCorner: Qt.CursorShape.SizeFDiagCursor,
            Qt.Corner.TopRightCorner: Qt.CursorShape.SizeBDiagCursor,
            Qt.Corner.BottomLeftCorner: Qt.CursorShape.SizeBDiagCursor,
            Qt.Corner.BottomRightCorner: Qt.CursorShape.SizeFDiagCursor,
        }.get(self._corner, Qt.CursorShape.ArrowCursor))

    def _move(self) -> None:
        if self.parent() is not None:
            rect: QRect = self.parent().rect()

            if self._corner == Qt.Corner.TopRightCorner:
                x, y = rect.topRight().toTuple()
                self.move(x - (self.width() + 5), y - 5)

            elif self._corner == Qt.Corner.TopLeftCorner:
                self.move(
                    rect.x(),
                    0
                )

            elif self._corner == Qt.Corner.BottomRightCorner:
                self.move(rect.right() - self.width() - 1, 
                          rect.bottom() - self.height() - 1)

            elif self._corner == Qt.Corner.BottomLeftCorner:
                self.move(0, rect.bottom() - self.height())

    def mousePressEvent(self, arg__1: QMouseEvent) -> None:
        self.parent().windowHandle().startSystemResize(qt_corner_to_edge(self._corner))
        return super().mousePressEvent(arg__1)

    def updateGeometry(self) -> None:
        self._move()
        return super().updateGeometry()


class FramelessWindow(WidgetBase, QMainWindow):
    startup_changed = Signal()

    def __init__(self, 
                 *,
                 title_bar: QWidget = None,
                 home: Union[QWidget, QLayout] = None,
                 status_bar: QStatusBar = None,
                 dock_widget: QDockWidget = None,
                 size_grip: QSizeGrip = None,
                 sys_tray: QSystemTrayIcon = None,
                 **kwargs):
        QMainWindow.__init__(self)
        super().__init__(**kwargs)

        self.__is_startup = False

        self._old_pos = self.pos()
        self._is_system_move = False
        self._size_grip = size_grip
        self._title_bar = title_bar

        self._set_central_widget()

        if self._title_bar is not None:
            self._set_title_bar(self._title_bar)
        else:
            self.mouseMoveEvent = self._mouseMoveEvent
            self.mousePressEvent = self._mousePressEvent

        if size_grip is not None:
            self._size_grip.setParent(self)

        if home is not None:
            self._set_home(home)

        self.setWindowFlags(
            self.windowFlags()
            | Qt.WindowType.FramelessWindowHint
        )

        self.setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, True)
        self.setAttribute(Qt.WidgetAttribute.WA_Hover, True)

        self.setAutoFillBackground(True)
        self.setUpdatesEnabled(True)
        self.setDocumentMode(True)
        self.setAnimated(True)

        if status_bar is not None:
            self._set_status_bar(status_bar)

        if dock_widget is not None:
            self._set_dock_widget(dock_widget)

        if sys_tray is not None:
            sys_tray.setParent(self)
            sys_tray.show()

    def closeEvent(self, event: QCloseEvent) -> None:
        super().closeEvent(event)
        self.__is_startup = False

    def showEvent(self, event: QShowEvent) -> None:
        super().showEvent(event)
        if not self.__is_startup:
            self.__is_startup = True
            self.startup_changed.emit()

    def _set_central_widget(self):
        self.__centralwidget = QWidget(self)
        self._vlayout = QVBoxLayout(self.__centralwidget)

        self.__centralwidget.setObjectName("__centralwidget")
        self.setCentralWidget(self.__centralwidget)

    def _set_home(self, home: QWidget):
        if isinstance(home, QLayout):
            home.setProperty("parent", 
                             self.__centralwidget)
            self._vlayout.addLayout(home)

        elif isinstance(home, QWidget):
            home.setParent(self)
            self._vlayout.addWidget(home)

        else:
            raise ValueError("the 'home' argument must be one of \
                the QWidget or QLayout instance.")

    def _set_title_bar(self, title_bar: QWidget):
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, 
                                 QSizePolicy.Policy.Fixed)

        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(title_bar.sizePolicy().hasHeightForWidth())
        title_bar.setSizePolicy(sizePolicy)

        title_bar.mouseMoveEvent = self._mouseMoveEvent
        title_bar.mousePressEvent = self._mousePressEvent

        self._vlayout.addWidget(title_bar)

    def resizeEvent(self, event: QResizeEvent):
        self._size_grip.updateGeometry()
        if self._title_bar is not None:
            self._title_bar.resize(self.width(), 
                                   self._title_bar.height())
        return QWidget.resizeEvent(self, event)

    def center_window(self, screen: QScreen = None):
        dst = screen if screen is not None else QApplication.primaryScreen()
        geo = self.frameGeometry()
        geo.moveCenter(QScreen.availableGeometry(dst).center())
        self.move(geo.center())

    def _mousePressEvent(self, event: QMouseEvent) -> None:
        if event.button() == event.buttons().LeftButton:
            self._old_pos = event.globalPos()
            self._is_system_move = self.windowHandle().startSystemMove()
        return super().mousePressEvent(event)

    def _mouseMoveEvent(self, event: QMouseEvent) -> None:
        if not self._is_system_move:
            delta = QPoint(event.globalPos() - self._old_pos)
            self.move(self.x() + delta.x(),
                      self.y() + delta.y())
            self._old_pos = event.globalPos()
        else:
            return super().mouseMoveEvent(event)

    def _set_status_bar(self, status_bar: QStatusBar) -> None:
        self.setStatusBar(status_bar)

    def _set_dock_widget(self, dock_widget: QDockWidget) -> None:
        self.addDockWidget(dock_widget)



from ..enums.widgets import Widgets, WidgetTypeVar
from ..enums.signals import Signals, SignalTypeVar
from ..enums.events import Events, EventTypeVar
from typing import Any, Callable, Sequence
from PySide6.QtCore import Qt, QObject
from PySide6.QtWidgets import QApplication
import caseconverter



class AbstractBase:
    def __init__(self, 
                 *,
                 uid: str = None,
                 signals: SignalTypeVar = None,
                 events: EventTypeVar = None,
                 methods: Sequence[tuple[str, Any]] = None,
                 **kwargs):

        self._set_uid(uid)
        self._set_events(events)
        self._set_signals(signals)
        self._set_methods(methods)
        self._set_property_from_kwargs(**kwargs)

    def _getattr(self, name: str, default: object = None) -> object:
        return getattr(self, name, default)

    def _set_uid(self, uid: str):
        if uid is not None:
            self.setObjectName(uid.strip())

    def _set_methods(self, methods):
        if not methods:
            return

        for method in methods:
            if ((func := self._getattr(method[0])) is not None
                and func.__class__.__name__ in (
                "builtin_function_or_method",
                "method_descriptor",
                "function"
                )):
                func(method[1])

    def _set_events(self, events: EventTypeVar):
        if not events:
            return

        for (event, slot) in events:
            try:
                ename = (caseconverter.camelcase(event.name)
                         if isinstance(event, Events)
                         else event.strip())

                ## self.{event} = newEventMethod
                self.__setattr__(ename, slot)
            except Exception as err:
                print("Error-Event: ", err)

    def _set_signals(self, 
                     signals: SignalTypeVar, 
                     disconnect: bool = False):

        if not signals:
            return

        for (signal, slot) in signals:
            try:
                sname = (caseconverter.camelcase(signal.name)
                         if isinstance(signal, Signals)
                         else signal.strip())

                if (attr := self._getattr(sname)) is not None:
                    if disconnect:
                        attr.disconnect(slot)
                    else:
                        attr.connect(slot)
                else:
                    ## print Warrning slot dose't found.
                    ...
            except Exception as err:
                print("Err-Signal: ", err)

    def _get_wtype(self, w_type: WidgetTypeVar = Widgets.any):
        return (w_type.value
                if isinstance(w_type, Widgets) 
                else w_type)

    def get(self, 
            uid: str, 
            w_type: WidgetTypeVar = Widgets.any) -> QObject:
        return self.findChild(self._get_wtype(w_type), 
                              uid.strip(), 
                              Qt.FindChildOption.FindChildrenRecursively)

    def _set_property_from_kwargs(self, **kwargs):
        for (name, value) in kwargs.items():
            if hasattr(self, name) or self.property(name) is not None:
                snake_case = caseconverter.snakecase(name)

                # handle signals
                if snake_case in Signals._member_names_:
                    self._getattr(name).connect(value)

                # handle events
                elif snake_case in Events._member_names_:
                    self.__setattr__(name, value)

                # handle parent parametar
                elif name == "parent":
                    if isinstance(value, str):
                        for parent in QApplication.topLevelWidgets():
                            if parent.objectName().strip() == value.strip():
                                self.setParent(parent)
                                break

                            elif (widget := parent.findChild(QObject, 
                                                             value.strip(), 
                                                             Qt.FindChildOption.FindChildrenRecursively)) is not None:
                                self.setParent(widget)
                                break

                    elif value.__class__.__name__ in (
                        "builtin_function_or_method",
                        "method_descriptor",
                        "function"
                    ):
                        self.setParent(value())
                    else:
                        self.setParent(value)

                # handle properties
                elif self.property(name) is not None:
                    self.setProperty(name, value)

                # handle set callables methods
                elif name.lower().startswith("set"):
                    if ((func := self._getattr(name)) is not None
                        and func.__class__.__name__ in (
                        "builtin_function_or_method",
                        "method_descriptor",
                        "function"
                      )):
                        func(value)


class NoneCheck:
    '''
    e.g
    NoneCheck(icon, self.setIcon)
    
    # the same!
    if icon is not None:
        self.setIcon()
    '''

    def __new__(cls, *args, **kwargs) -> Any:
        instance = super().__new__(cls)
        return instance.__init__(*args, **kwargs)

    def __init__(self, 
                 _callable: Callable,
                 value: Any) -> Any:

        if value is not None:
            return _callable(value)

## Declarative
class DuplicateKeyError(Exception):
    def __init__(self, uid: str):
        super().__init__(uid)
        print(f"'{uid}' was registered in this dictionary!")


class _TrackingDeclarative:
    def __init__(self):
        self._instances_dict = dict()

    def _register(self, uid: str, obj: Any) -> None:
        if not isinstance(uid, str):
            raise ValueError("invalid uid, must be str type.")

        if uid in self._instances_dict:
            raise DuplicateKeyError(uid)
        self._instances_dict[uid] = obj

    def _deregister(self, uid: str) -> Any:
        """ deregister widget from manager """
        if not isinstance(uid, str):
            raise ValueError("invalid uid, must be str type.")

        if uid not in self._instances_dict:
            return

        return self._instances_dict.pop(uid)

    def pop(self, uid: str) -> Any:
       self._deregister(uid)

    def get(self, uid: str) -> Any:
        return self._instances_dict.get(uid)

    def items(self) -> Any:
        return self._instances_dict.items()

TrackingDeclarative = _TrackingDeclarative()


class BehaviorDeclarative:
    """
    `__init__` method can now return a value

    ### usage example
    ```python
    class Object(BehaviorDeclarative):
        def __init__(self, *args, **kwargs):
            return ...
    ```
    
    #### NOTE: use `uid` parameter to store this class in `TrackingDeclarative`, \
        and call it with Api.fetch
    """

    def __new__(cls, *args, **kwargs) -> Any:
        instance = super().__new__(cls)
        _uid = (kwargs.pop
                if 'uid' not in
                instance.__init__.__code__.co_varnames
                else kwargs.get)("uid", None)

        if _uid is not None:
            TrackingDeclarative._register(_uid, instance)

        return instance.__init__(*args, **kwargs)




from PySide6.QtWidgets import QFrame
from ..core import WidgetBase


class _Line(WidgetBase, QFrame):
    def __init__(self,
                 shadow: QFrame.Shadow = QFrame.Shadow.Sunken,
                 **kwargs):
        QFrame.__init__(self)
        super().__init__(**kwargs)

        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setFrameShadow(shadow)


class HLine(_Line):
    def __init__(self,
                 *,
                 shadow: QFrame.Shadow = QFrame.Shadow.Sunken,
                 **kwargs):
        super().__init__(shadow=shadow, **kwargs)

        self.setFrameShape(QFrame.Shape.HLine)


class VLine(_Line):
    def __init__(self, 
                 *,
                 shadow: QFrame.Shadow = QFrame.Shadow.Sunken,
                 **kwargs):
        super().__init__(shadow=shadow,**kwargs)

        self.setFrameShape(QFrame.Shape.VLine)


#!/usr/bin/python3

from PySide6.QtGui import QBrush, QColor, QPaintEvent, QPainter, QPen
from PySide6.QtCore import QPointF, QRectF, QSize, Qt
from PySide6.QtWidgets import QWidget
from ...core import PainterBase


class StatusEdgePaint(PainterBase):
    def __init__(self,
                 *,
                 child: QWidget,
                 corner: Qt.Corner = Qt.Corner.BottomRightCorner,
                 status_color: QColor = None,
                 border_color: QColor = None,
                 padding: float = 1.5,
                 width: int = None
                 ) -> QWidget:
        super().__init__(child)

        self._padding = padding
        self._corner = corner

        self._status_color = (
            status_color 
            if status_color is not None 
            else QColor("white")
        )

        self._border_color = (
            border_color 
            if border_color is not None 
            else QColor("#151617")
        )

        self._width = (
            width 
            if width is not None
            else 13
        )

        return self._parent

    def set_status_color(self, color: QColor):
        self._status_color = color
        self.update()

    def update_state(self, color: QColor):
        self._status_color = color
        self.update()

    def _paint(self, event: QPaintEvent):
        # self._parent.__class__.paintEvent(self._parent, event)
        self._super_paintEvent(event)

        painter = QPainter(self._parent)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        pen = QPen()
        pen.setWidth(3)
        pen.setColor(self._border_color)
        painter.setPen(pen)

        # BRUSH/STATUS COLOR
        painter.setBrush(QBrush(QColor(self._status_color)))

        # DRAW
        if self._corner == Qt.Corner.BottomRightCorner:
            point = QPointF((self._parent.width() - self._width) - self._padding,
                            (self._parent.height() - self._width) - self._padding)
        elif self._corner == Qt.Corner.BottomLeftCorner:
            point = QPointF(self._padding, (self._parent.height() - self._width) - self._padding)
        elif self._corner == Qt.Corner.TopRightCorner:
            point = QPointF((self._parent.width() - self._width) - self._padding, self._padding)
        elif self._corner == Qt.Corner.TopLeftCorner:
            point = QPointF(self._padding, self._padding)

        painter.drawEllipse(QRectF(point, QSize(self._width, self._width)))


from .drop_shadow import DropShadowEffect
from .colorize import ColorizeEffect
from .opacity import OpacityEffect
from .blur import BlurEffect
from .neumorphism import (
    InsideNeumorphismEffect, 
    OutsideNeumorphismEffect
)


from typing import Union
from PySide6.QtWidgets import QStackedWidget, QWidget
from ...utils._routes import Routes
from ...core import WidgetBase


class StackedWidget(WidgetBase, QStackedWidget):
    def __init__(self, 
                 *,
                 children: Union[list[QWidget], dict[str, QWidget]] = None,
                 **kwargs):
        QStackedWidget.__init__(self)
        super().__init__(**kwargs)

        if isinstance(children, dict):
            self._routes = Routes(self, "/")
            self.__setattr__("route", self._routes)
            self.__setattr__("push", lambda route: self._routes.push(route))

        if children is not None:
            self._set_children(children)

    def _set_children(self, children) -> None:
        if not isinstance(children, dict):
            for child in children:
                self.addWidget(child)
        else:
            for route, child in children.items():
                if child is not None:
                    self._routes.add(route, child)


import re
import os
import platform
import socket
import sys
import webbrowser
from pathlib import Path
from .overload import staticproperty


class SystemType:
    @staticproperty
    def is_linux() -> bool:
        return platform.system() == "Linux"

    @staticproperty
    def is_linux_server() -> bool:
        if platform.system() == "Linux":
            # check if it's WSL
            p = "/proc/version"
            if os.path.exists(p):
                with open(p, "r") as file:
                    if "microsoft" in file.read():
                        return False  # it's WSL, not a server
            return os.environ.get("XDG_CURRENT_DESKTOP") is None
        return False

    @staticproperty
    def is_windows() -> bool:
        return platform.system() == "Windows"

    @staticproperty
    def is_macos() -> bool:
        return platform.system() == "Darwin"

    @staticproperty
    def is_java() -> bool:
        return platform.system() == "Java"


def get_platform():
    if (p := platform.system().lower()):
        return p

    raise Exception(f"Unsupported platform: {p}")


def get_arch():
    a = platform.machine().lower()
    if a == "x86_64" or a == "amd64":
        return "amd64"
    elif a == "arm64" or a == "aarch64":
        return "arm64"
    elif a.startswith("arm"):
        return "arm_7"
    else:
        raise Exception(f"Unsupported architecture: {a}")


def open_in_browser(url):
    webbrowser.open(url)


# https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def which(program, exclude_exe=None):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    for path in os.environ["PATH"].split(os.pathsep):
        exe_file = os.path.join(path, program)
        if is_exe(exe_file) and (
            exclude_exe is None
            or (exclude_exe is not None and exclude_exe.lower() != exe_file.lower())
        ):
            return exe_file

    return None


def is_within_directory(directory, target):
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def safe_tar_extractall(tar, path=".", members=None, *, numeric_owner=False):
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner=numeric_owner)


def is_localhost_url(url):
    return (
        "://localhost/" in url
        or "://localhost:" in url
        or "://127.0.0.1/" in url
        or "://127.0.0.1:" in url
    )


def get_package_root_dir():
    return str(Path(__file__).parent)


def get_package_bin_dir():
    return os.path.join(get_package_root_dir(), "bin")


def get_package_web_dir():
    web_root_dir = os.environ.get("FLET_WEB_PATH")
    return web_root_dir if web_root_dir else os.path.join(get_package_root_dir(), "web")


def get_free_tcp_port():
    sock = socket.socket()
    sock.bind(("", 0))
    return sock.getsockname()[1]


def get_current_script_dir():
    pathname = os.path.dirname(sys.argv[0])
    return os.path.abspath(pathname)


def get_pi_version():
    """Detect the version of the Raspberry Pi by reading the revision field value from '/proc/cpuinfo'
    See: https://www.raspberrypi.org/documentation/hardware/raspberrypi/revision-codes/README.md
    Based on: https://github.com/adafruit/Adafruit_Python_GPIO/blob/master/Adafruit_GPIO/Platform.py
    """  # noqa
    # Check if file exist
    if not os.path.isfile('/proc/cpuinfo'):
        return None

    with open('/proc/cpuinfo', 'r') as f:
        cpuinfo = f.read()

    # Match a line like 'Revision   : a01041'
    revision = re.search(r'^Revision\s+:\s+(\w+)$', 
                         cpuinfo, 
                         flags=re.MULTILINE | re.IGNORECASE)
    if not revision:
        # Couldn't find the hardware revision, assume it is not a Pi
        return None

    # Determine the Pi version using the processor bits using the new-style
    # revision format
    revision = int(revision.group(1), base=16)
    if revision & 0x800000:
        return ((revision & 0xF000) >> 12) + 1

    # If it is not using the new style revision format,
    # then it must be a Raspberry Pi 1
    return 1


def is_32bit():
    return sys.maxsize <= 2**32


def is_64bit():
    return sys.maxsize > 2**32


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os


class ProcessInfo(object):

    def __init__(self):
        self.update()

    def update(self):
        processes = [int(entry) for entry in os.listdir("/proc") if entry.isdigit()]
        parent = {}
        children = {}
        commands = {}
        for pid in processes:
            try:
                f = open("/proc/%s/stat" % pid)
                # print(pid)
            except IOError:
                continue
            stat = f.read().split()
            f.close()
            cmd = stat[1]
            try:
                ppid = int(stat[3])
            except ValueError:
                pass
            parent[pid] = ppid
            children.setdefault(ppid, []).append(pid)
            commands[pid] = cmd
        self.parent = parent
        self.children = children
        self.commands = commands

    def all_children(self, pid):
        cl = self.children.get(pid, [])[:]
        for child_pid in cl:
            cl.extend(self.children.get(child_pid, []))
        return cl

    def dump(self, pid, _depth=0):
        print(" " * (_depth * 2), pid, self.commands[pid])
        for child_pid in self.children.get(pid, []):
            self.dump(child_pid, _depth + 1)

    def cwd(self, pid):
        try:
            path = os.readlink("/proc/%s/cwd" % pid)
        except OSError:
            return
        return path


if __name__ == "__main__":
    pi = ProcessInfo()
    pi.dump(4984)
    print(pi.all_children(4984))
    print(pi.cwd(4984))
    print(pi.cwd(pi.all_children(4984)[-1]))



from PySide6.QtCore import Signal
from PySide6.QtWidgets import QLabel
from ..core import WidgetBase


class Label(WidgetBase, QLabel):
    textChanged = Signal(str)

    def __init__(self, **kwargs):
        QLabel.__init__(self)
        super().__init__(**kwargs)

    def setText(self, arg__1: str) -> None:
        super().setText(arg__1)
        self.textChanged.emit(arg__1)


from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QFontMetrics
from ..core.base import BehaviorDeclarative


class ElidedText(BehaviorDeclarative):
    """
    ElidedText(
        child = Label("Your text here"),
        width = Api("window").width(),
        elided_mode = Qt.TextElideMode.ElideRight
    )
    """
    def __init__(self, 
                 *,
                 child: QWidget,
                 width: int = None,
                 elide_mode: Qt.TextElideMode = Qt.TextElideMode.ElideRight,
                 **kwargs) -> QWidget:

        if not hasattr(child, "setText"):
            raise ValueError("invalid child widget!")

        if hasattr(child, "textChanged"):
            child.textChanged.connect(lambda text: self._set_elided(child, width, elide_mode))

        self._set_elided(child, width, elide_mode)

        return child

    def _set_elided(self, child, width: int, elide_mode):
        width = width if width is not None else self._get_parent_width(child)

        child.blockSignals(True)
        child.setText(child.fontMetrics().elidedText(child.text(), elide_mode, width))
        child.blockSignals(False)

    def _get_parent_width(self, child: QWidget) -> int:
        parent = child.topLevelWidget()
        if parent is None:
            parent = child.parentWidget()

        if parent is not None:
            return parent.width()

        return child.width()

    @staticmethod
    def elided(text: str,
               width: int,
               elide_mode: Qt.TextElideMode = Qt.TextElideMode.ElideRight) -> str:
        return QFontMetrics(text).elidedText(text, elide_mode, width)



from typing import Callable, Union
from PySide6.QtGui import QAction, QIcon, QPixmap
from Qtica import BehaviorDeclarative


class MenuSeparator:
    ...


class MenuSection(BehaviorDeclarative):
    def __init__(self,
                 text: str,
                 icon: Union[QIcon, QPixmap] = None):
        return (text, icon) if icon is not None else (text,)


class MenuSimpleAction(BehaviorDeclarative):
    def __init__(self, 
                 text: str = None,
                 icon: Union[QIcon, QPixmap] = None, 
                 callable: Callable = None):

        action = QAction()

        if icon is not None:
            action.setIcon(icon)

        if callable is not None:
            action.triggered.connect(callable)

        action.setText(text)
        return action


#!/usr/bin/python3

from __future__ import division


__all__ = ["Contrast"]


class Contrast:
    def __init__(self):
        ...

    def rgb(self, 
            rgb1: tuple[int, int, int], 
            rgb2: tuple[int, int, int]) -> float:

        for r, g, b in (rgb1, rgb2):
            if not 0.0 <= r <= 1.0:
                raise ValueError("r is out of valid range (0.0 - 1.0)")
            if not 0.0 <= g <= 1.0:
                raise ValueError("g is out of valid range (0.0 - 1.0)")
            if not 0.0 <= b <= 1.0:
                raise ValueError("b is out of valid range (0.0 - 1.0)")

        l1 = self._relative_luminance(*rgb1)
        l2 = self._relative_luminance(*rgb2)

        if l1 > l2:
            return (l1 + 0.05) / (l2 + 0.05)
        else:
            return (l2 + 0.05) / (l1 + 0.05)

    def _relative_luminance(self, r, g, b) -> float:
        r = self._linearize(r)
        g = self._linearize(g)
        b = self._linearize(b)

        return 0.2126 * r + 0.7152 * g + 0.0722 * b

    def _linearize(self, v) -> float:
        if v <= 0.03928:
            return v / 12.92
        else:
            return ((v + 0.055) / 1.055) ** 2.4

    def passes_AA(self, contrast: float, large=False) -> bool:
        if large:
            return contrast >= 3.0
        else:
            return contrast >= 4.5

    def passes_AAA(self, contrast: float, large=False) -> bool:
        if large:
            return contrast >= 4.5
        else:
            return contrast >= 7.0


class ContrastFont:
    @classmethod
    def color(cls, rgb: tuple[int, int, int]) -> tuple[int, int, int]:
        '''
        Calculate the perceptive luminance (aka luma) - human eye favors green color... 
        '''
        r, g, b, *_ = rgb
        luma = ((0.299 * r) + (0.587 * g) + (0.114 * b)) / 255
        return (0,) * 3 if luma > 0.5 else (255,) * 3


# Qtica
__version__ = "0.2.0"
__version_info__ = (0, 2, 0, "", "")

# Python
__minimum_python_version__ = (3, 7)
__maximum_python_version__ = (3, 11)

# PySide6
__pyside_version__ = "6.5.0"

# base classes
from .core import (
    AbstractBase, 
    QObjectBase,
    WidgetBase, 
    NoneCheck,

    # declarative
    BehaviorDeclarative, 
    WidgetDeclarative, 
    QObjectDeclarative,
    DuplicateKeyError,
    TrackingDeclarative,

    # platforms
    AndroidInit,
    IOSInit,
    LinuxInit,
    MacOSInit,
    WindowsInit,

    Api,
    QStyleSheet,
    AbstractTool,
    ToolBase,
    AbstractIcon,
    IconBase,
    AbstractPainter,
    PainterBase
)


#!/usr/bin/python3

from PySide6.QtWidgets import QGridLayout, QApplication, QLabel, QDialog, QWidget
from PySide6.QtGui import QFont, QMouseEvent
from PySide6.QtCore import QTimer, Qt


class _DisplayLargText(QDialog):
    def __init__(self,
                 parent = None,
                 close_on_mouse_press: bool = True) -> None:
        QDialog.__init__(self, parent)

        self._close_on_mouse_press = close_on_mouse_press

        self.setWindowFlags(
            self.windowFlags() 
            | Qt.WindowType.FramelessWindowHint)

        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground, True)
        self.setAttribute(Qt.WidgetAttribute.WA_AlwaysStackOnTop, True)
        self.setWindowOpacity(0.9)

        self.setStyleSheet("""
        QDialog, QWidget{
            padding: 8px;
            border-radius: 8px;
            background-color: #0d0d0d;
            color: #ffffff;
        }
        """)

        self.form = QWidget(self)
        self.grid_layout1 = QGridLayout(self)
        self.grid_layout2 = QGridLayout(self.form)
        self.text = QLabel(self.form)

        self.text.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.text.setWordWrap(True)

        self.grid_layout2.addWidget(self.text, 0, 0, 0, 0, Qt.AlignmentFlag.AlignCenter)

        self.grid_layout1.addWidget(self.form)
        self.form.setLayout(self.grid_layout2)

        self.setLayout(self.grid_layout1)

    def start_closeing(self):
        self.close()

    def start_show(self):
        if self.isHidden():
            self.show()
            self.adjustSize()

    def display_text(self, 
                     text: str, 
                     font: QFont = None, 
                     timeout: int = 2000) -> None:
        self.text.setText(text)
        if font is not None:
            self.text.setFont(font)
        else:
            font = self.font()
            font.setBold(True)
            font.setPixelSize(50)
            self.text.setFont(font)

        QTimer.singleShot(0, self.start_show)
        QTimer.singleShot(timeout, self.start_closeing)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if self._close_on_mouse_press:
            self.start_closeing()
        return super().mousePressEvent()


class LargTextDialog:
    @classmethod
    def display(cls,
                *,
                text: str,
                font: QFont = None,
                timeout: int = 2000,
                auto_close: bool = False,
                parent: QWidget = None,
                **kwargs):

        if not parent:
            parent = QApplication.activeWindow()

        larg_text_dialog = _DisplayLargText(parent, auto_close, **kwargs)
        larg_text_dialog.display_text(text, font, timeout)



from enum import IntEnum, auto


class Positions(IntEnum):
    """
    left   => West
    top    => North
    bottom => South
    right  => East
    center => Center
    """

    left = auto()
    top = auto()
    bottom = auto()
    right = auto()
    center = auto()

    west = left
    north = top
    south = bottom
    east = right



# coding:utf-8
from collections import deque
from math import cos, pi

from PySide6.QtCore import QDateTime, Qt, QTimer, QPoint
from PySide6.QtGui import QWheelEvent
from PySide6.QtWidgets import QApplication, QScrollArea, QAbstractScrollArea
from ..core.base import BehaviorDeclarative
from ..enums.smooth_scroll import SmoothMode


class _SmoothScroll:
    """ Scroll smoothly """

    def __init__(self,
                 widget: QScrollArea, 
                 orient: Qt.Orientation = Qt.Orientation.Vertical):
        """
        Parameters
        ----------
        widget: QScrollArea
            scroll area to scroll smoothly

        orient: Orientation
            scroll orientation
        """
        self.widget = widget
        self.orient = orient
        self.fps = 60
        self.duration = 400
        self.stepsTotal = 0
        self.stepRatio = 1.5
        self.acceleration = 1
        self.lastWheelEvent = None
        self.scrollStamps = deque()
        self.stepsLeftQueue = deque()
        self.smoothMoveTimer = QTimer(widget)
        self.smoothMode = SmoothMode(SmoothMode.linear)
        self.smoothMoveTimer.timeout.connect(self.__smoothMove)

    def setSmoothMode(self, smoothMode):
        """ set smooth mode """
        self.smoothMode = smoothMode

    def wheelEvent(self, e):
        # only process the wheel events triggered by mouse, fixes issue #75
        delta = e.angleDelta().y() if e.angleDelta().y() != 0 else e.angleDelta().x()
        if self.smoothMode == SmoothMode.normal or abs(delta) % 120 != 0:
            QAbstractScrollArea.wheelEvent(self.widget, e)
            return

        # push current time to queque
        now = QDateTime.currentDateTime().toMSecsSinceEpoch()
        self.scrollStamps.append(now)
        while now - self.scrollStamps[0] > 500:
            self.scrollStamps.popleft()

        # adjust the acceration ratio based on unprocessed events
        accerationRatio = min(len(self.scrollStamps) / 15, 1)
        self.lastWheelPos = e.position()
        self.lastWheelGlobalPos = e.globalPosition()

        # get the number of steps
        self.stepsTotal = self.fps * self.duration / 1000

        # get the moving distance corresponding to each event
        delta = delta* self.stepRatio
        if self.acceleration > 0:
            delta += delta * self.acceleration * accerationRatio

        # form a list of moving distances and steps, and insert it into the queue for processing.
        self.stepsLeftQueue.append([delta, self.stepsTotal])

        # overflow time of timer: 1000ms/frames
        self.smoothMoveTimer.start(int(1000 / self.fps))

    def __smoothMove(self):
        """ scroll smoothly when timer time out """
        totalDelta = 0

        # Calculate the scrolling distance of all unprocessed events,
        # the timer will reduce the number of steps by 1 each time it overflows.
        for i in self.stepsLeftQueue:
            totalDelta += self.__subDelta(i[0], i[1])
            i[1] -= 1

        # If the event has been processed, move it out of the queue
        while self.stepsLeftQueue and self.stepsLeftQueue[0][1] == 0:
            self.stepsLeftQueue.popleft()

        # construct wheel event
        if self.orient == Qt.Vertical:
            pixelDelta = QPoint(round(totalDelta), 0)
            bar = self.widget.verticalScrollBar()
        else:
            pixelDelta = QPoint(0, round(totalDelta))
            bar = self.widget.horizontalScrollBar()

        e = QWheelEvent(
            self.lastWheelPos,
            self.lastWheelGlobalPos,
            pixelDelta,
            QPoint(round(totalDelta), 0),
            Qt.MouseButton.LeftButton,
            Qt.KeyboardModifier.NoModifier,
            Qt.ScrollPhase.ScrollBegin,
            False,
        )

        # send wheel event to app
        QApplication.sendEvent(bar, e)

        # stop scrolling if the queque is empty
        if not self.stepsLeftQueue:
            self.smoothMoveTimer.stop()

    def __subDelta(self, delta, stepsLeft):
        """ get the interpolation for each step """
        m = self.stepsTotal / 2
        x = abs(self.stepsTotal - stepsLeft - m)

        res = 0
        if self.smoothMode == SmoothMode.normal:
            res = 0
        elif self.smoothMode == SmoothMode.constant:
            res = delta / self.stepsTotal
        elif self.smoothMode == SmoothMode.linear:
            res = 2 * delta / self.stepsTotal * (m - x) / m
        elif self.smoothMode == SmoothMode.quadrati:
            res = 3 / 4 / m * (1 - x * x / m / m) * delta
        elif self.smoothMode == SmoothMode.cosine:
            res = (cos(x * pi / m) + 1) / (2 * m) * delta

        return res


class SmoothScroll(BehaviorDeclarative):
    def __init__(self, 
                 *,
                 child: QScrollArea,
                 orient: Qt.Orientation = Qt.Orientation.Vertical) -> QScrollArea:
        self.smooth_scroll = _SmoothScroll(child, orient)
        return child


#!/usr/bin/python3

from typing import Any, Sequence
from .base import BehaviorDeclarative


class AbstractTool(BehaviorDeclarative):
    def __init__(self, 
                 *,
                 methods: Sequence[tuple[str, Any]] = None,
                 **kwargs):

        self._set_methods(methods)
        self._set_property_from_kwargs(**kwargs)

    def _set_methods(self, methods):
        if not methods:
            return

        for method in methods:
            if ((func := self._getattr(method[0])) is not None
                and func.__class__.__name__ in (
                "builtin_function_or_method",
                "method_descriptor",
                "function"
                )):
                func(method[1])

    def _set_property_from_kwargs(self, **kwargs):
        for (name, value) in kwargs.items():
            if hasattr(self, name):
                # handle set callables methods
                if name.lower().startswith("set"):
                    if ((func := self._getattr(name)) is not None
                        and func.__class__.__name__ in (
                        "builtin_function_or_method",
                        "method_descriptor",
                        "function"
                      )):
                        func(value)


class ToolBase(AbstractTool):
    pass


#!/usr/bin/python3

from typing import Union
from PySide6.QtGui import QColor, QIcon, QIconEngine, QImage, QPaintEvent, QPainter, QPixmap
from PySide6.QtWidgets import QWidget
from PySide6.QtCore import QSize, Qt
from ..core import AbstractIcon
from ..core import WidgetBase
from ..tools.icon import Icon


class IconWidget(WidgetBase, QWidget):
    def __init__(self,
                 *,
                icon: Union[str,
                            QIcon,
                            QIconEngine,
                            QPixmap,
                            QImage,
                            AbstractIcon],
                alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter | Qt.AlignmentFlag.AlignHCenter,
                mode: QIcon.Mode = QIcon.Mode.Active,
                state: QIcon.State = QIcon.State.On,
                color: QColor = None,
                size: Union[QSize, tuple, int] = None,
                **kwargs):
        QWidget.__init__(self)
        super().__init__(**kwargs)

        self._icon = Icon(icon, color, size)
        self._alignment = alignment
        self._mode = mode
        self._state = state

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing 
                               | QPainter.RenderHint.SmoothPixmapTransform
                               | QPainter.RenderHint.TextAntialiasing)

        self._icon.paint(painter, 
                         self.rect(),
                         self._alignment, 
                         self._mode, 
                         self._state)


from enum import Enum


class EnvVars(Enum):
    enable_highdpi_scaling: str = "QT_ENABLE_HIGHDPI_SCALING"
    scale_factor: str = "QT_SCALE_FACTOR"
    quick_controls_style: str = "QT_QUICK_CONTROLS_STYLE"
    quick_controls_fallback_style: str = "QT_QUICK_CONTROLS_FALLBACK_STYLE"
    quick_controls_conf: str = "QT_QUICK_CONTROLS_CONF"
    quick_controls_hover_enabled: str = "QT_QUICK_CONTROLS_HOVER_ENABLED"
    quick_controls_imagine_path: str = "QT_QUICK_CONTROLS_IMAGINE_PATH"
    quick_controls_imagine_smooth: str = "QT_QUICK_CONTROLS_IMAGINE_SMOOTH"
    quick_controls_material_theme: str = "QT_QUICK_CONTROLS_MATERIAL_THEME"
    quick_controls_material_variant: str = "QT_QUICK_CONTROLS_MATERIAL_VARIANT"
    quick_controls_material_accent: str = "QT_QUICK_CONTROLS_MATERIAL_ACCENT"
    quick_controls_material_primary: str = "QT_QUICK_CONTROLS_MATERIAL_PRIMARY"
    quick_controls_material_foreground: str = "QT_QUICK_CONTROLS_MATERIAL_FOREGROUND"
    quick_controls_material_background: str = "QT_QUICK_CONTROLS_MATERIAL_BACKGROUND"
    quick_controls_universal_theme: str = "QT_QUICK_CONTROLS_UNIVERSAL_THEME"
    quick_controls_universal_accent: str = "QT_QUICK_CONTROLS_UNIVERSAL_ACCENT"
    quick_controls_universal_foreground: str = "QT_QUICK_CONTROLS_UNIVERSAL_FOREGROUND"
    quick_controls_universal_background: str = "QT_QUICK_CONTROLS_UNIVERSAL_BACKGROUND"
    qpa_platform: str = "QT_QPA_PLATFORM"
    im_module: str = "QT_IM_MODULE"


from typing import Callable
from pynput import mouse
from PIL import ImageGrab
from PySide6.QtCore import Signal, QObject
from PySide6.QtGui import QColor
from PySide6.QtWidgets import QApplication
from .icon import Icon



class ColorPicker(QObject):
    color_changed = Signal(QColor)

    def __init__(self,
                 *,
                 mouse_icon: Icon = None,
                 color_changed: Callable = None):
        super().__init__()

        self._mouse_icon = mouse_icon
        self._is_running = False
        self._on_color_changed = color_changed
        self._mouse_listener = mouse.Listener(on_click = self._on_click)
        self._mouse_listener.start()

        if color_changed is not None:
            self.color_changed.connect(color_changed)

    def start(self):
        self._is_running = True
        if self._mouse_icon is not None:
            QApplication.setOverrideCursor(
                self._mouse_icon.pixmap(32, 32)
            )

    def close(self) -> None:
        self._is_running = False
        if self._mouse_icon is not None:
            QApplication.restoreOverrideCursor()
        self._mouse_listener.stop()

    def _get_color(self, x: int, y: int):
        im = ImageGrab.grab(bbox=(x, y, x + 1, y + 1))
        rgbim = im.convert('RGB')
        r,g,b,*_ = rgbim.getpixel((0,0))
        return r, g, b

    def _on_click(self, x, y, button, pressed):
        if not self._is_running:
            return

        if pressed and button == mouse.Button.left:
            self.color_changed.emit(QColor(*self._get_color(x, y)))
            self._is_running = False
            QApplication.restoreOverrideCursor()


# coding:utf-8

from PySide6.QtWidgets import QLayout, QLayoutItem, QWidget
from PySide6.QtCore import (
    QSize, 
    QPoint, 
    Qt, 
    QRect, 
    QPropertyAnimation, 
    QParallelAnimationGroup, 
    QEasingCurve
)

from typing import Union
from ..core import QObjectBase


class _FlowLayout(QLayout):
    """ Flow layout """

    def __init__(self, 
                 parent=None, 
                 enable_animation: bool = False, 
                 enable_tight: bool = False):
        """
        Parameters
        ----------
        parent:
            parent window or layout

        enable_animation: bool
            whether to add moving animation

        enable_tight: bool
            whether to use the tight layout when widgets are hidden
        """
        super().__init__(parent)
        self._items: list[QLayoutItem] = []
        self._anis = []
        self._aniGroup = QParallelAnimationGroup(self)
        self._verticalSpacing = 10
        self._horizontalSpacing = 10
        self.enable_animation = enable_animation
        self.enable_tight = enable_tight

    def addItem(self, item):
        self._items.append(item)

    def addWidget(self, w):
        super().addWidget(w)
        if not self.enable_animation:
            return

        ani = QPropertyAnimation(w, b'geometry')
        ani.setEndValue(QRect(QPoint(0, 0), w.size()))
        ani.setDuration(300)
        w.setProperty('flowAni', ani)
        self._anis.append(ani)
        self._aniGroup.addAnimation(ani)

    def setAnimation(self, duration, ease=QEasingCurve.Linear):
        """ set the moving animation

        Parameters
        ----------
        duration: int
            the duration of animation in milliseconds

        ease: QEasingCurve
            the easing curve of animation
        """
        if not self.enable_animation:
            return

        for ani in self._anis:
            ani.setDuration(duration)
            ani.setEasingCurve(ease)

    def count(self):
        return len(self._items)

    def itemAt(self, index: int):
        if 0 <= index < len(self._items):
            return self._items[index]

        return None

    def takeAt(self, index: int):
        if 0 <= index < len(self._items):
            item = self._items[index]
            ani = item.widget().property('flowAni')
            if ani:
                self._anis.remove(ani)
                self._aniGroup.removeAnimation(ani)
                ani.deleteLater()

            return self._items.pop(index).widget()

        return None

    def removeWidget(self, widget):
        for i, item in enumerate(self._items):
            if item.widget() is widget:
                return self.takeAt(i)

    def removeAllWidgets(self):
        """ remove all widgets from layout """
        while self._items:
            self.takeAt(0)

    def takeAllWidgets(self):
        """ remove all widgets from layout and delete them """
        while self._items:
            w = self.takeAt(0)
            if w:
                w.deleteLater()

    def expandingDirections(self):
        return Qt.Orientation(0)

    def hasHeightForWidth(self):
        return True

    def heightForWidth(self, width: int):
        """ get the minimal height according to width """
        return self._doLayout(QRect(0, 0, width, 0), False)

    def setGeometry(self, rect: QRect):
        super().setGeometry(rect)
        self._doLayout(rect, True)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self._items:
            size = size.expandedTo(item.minimumSize())

        m = self.contentsMargins()
        size += QSize(m.left()+m.right(), m.top()+m.bottom())

        return size

    def setVerticalSpacing(self, spacing: int):
        """ set vertical spacing between widgets """
        self._verticalSpacing = spacing

    def verticalSpacing(self):
        """ get vertical spacing between widgets """
        return self._verticalSpacing

    def setHorizontalSpacing(self, spacing: int):
        """ set horizontal spacing between widgets """
        self._horizontalSpacing = spacing

    def horizontalSpacing(self):
        """ get horizontal spacing between widgets """
        return self._horizontalSpacing

    def _doLayout(self, rect: QRect, move: bool):
        """ adjust widgets position according to the window size """
        margin = self.contentsMargins()
        x = rect.x() + margin.left()
        y = rect.y() + margin.top()
        rowHeight = 0
        spaceX = self.horizontalSpacing()
        spaceY = self.verticalSpacing()

        for i, item in enumerate(self._items):
            if item.widget() and not item.widget().isVisible() and self.enable_tight:
                continue

            nextX = x + item.sizeHint().width() + spaceX

            if nextX - spaceX > rect.right() and rowHeight > 0:
                x = rect.x() + margin.left()
                y = y + rowHeight + spaceY
                nextX = x + item.sizeHint().width() + spaceX
                rowHeight = 0

            if move:
                if not self.enable_animation:
                    item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))
                else:
                    self._anis[i].stop()
                    self._anis[i].setEndValue(QRect(QPoint(x, y), item.sizeHint()))

            x = nextX
            rowHeight = max(rowHeight, item.sizeHint().height())

        if self.enable_animation:
            self._aniGroup.stop()
            self._aniGroup.start()

        return y + rowHeight - rect.y()


class FlowLayout(QObjectBase, _FlowLayout):
    def __init__(self,
                 *,
                 children: list[Union[QWidget, QLayoutItem]] = None,
                 enable_animation: bool = False, 
                 enable_tight: bool = False,
                 duration: float = None,
                 easing_curve: QEasingCurve.Type = None,
                 **kwargs) -> None:
        _FlowLayout.__init__(self, 
                            enable_animation=enable_animation, 
                            enable_tight=enable_tight)
        super().__init__(**kwargs)

        self._set_animation(duration, 
                            easing_curve)

        if children is not None:
            for child in children:
                if isinstance(child, QLayoutItem):
                    self.addItem(child)

                elif isinstance(child, QWidget):
                    self.addWidget(child)

    def _set_animation(self, duration, easing_curve) -> None:
        if not self.enable_animation:
            return

        if duration is not None or easing_curve is not None:
            for ani in self._anis:
                if duration is not None:
                    ani.setDuration(duration)
                if easing_curve is not None:
                    ani.setEasingCurve(easing_curve)


#!/usr/bin/python3

from typing import Union
from PySide6.QtWidgets import QStackedWidget, QWidget
from PySide6.QtCore import (
    Qt,
    QEasingCurve,
    QPoint,
    Slot,
    QParallelAnimationGroup,
    QPropertyAnimation,
    QAbstractAnimation,
)
from ...core import WidgetBase
from ...utils._routes import Routes



class _SlidingStackedWidget(QStackedWidget):
    def __init__(
        self,
        parent: object = None,
        direction: Qt.Orientation = Qt.Orientation.Horizontal,
        animation_type: QEasingCurve = QEasingCurve.Type.OutCubic,
        speed: int = 1000,
        current: int = 0,
        next: int = 0,
        wrap: bool = False,
        pnow=None,
        active: bool = False
    ):
        QStackedWidget.__init__(self, parent)

        self.m_direction = direction  # QtCore.Qt.Vertical
        self.m_speed = speed
        self.m_animationtype = animation_type
        self.m_now = current
        self.m_next = next
        self.m_wrap = wrap
        self.m_pnow = pnow
        self.m_active = active

    def setDirection(self, direction: Qt.Orientation):
        self.m_direction = direction

    def setSpeed(self, speed: int):
        self.m_speed = speed

    def setAnimation(self, animationtype: QEasingCurve):
        self.m_animationtype = animationtype

    def setWrap(self, wrap: bool):
        self.m_wrap = wrap

    def setCurrentIndex(self, index: int) -> None:
        self.slideInIdx(index)
        return super().setCurrentIndex(index)
    
    def setCurrentWidget(self, w: QWidget) -> None:
        self.slideInWgt(w)
        return super().setCurrentWidget(w)

    @Slot()
    def slideInPrev(self):
        now = self.currentIndex()
        if self.m_wrap or now > 0:
            self.slideInIdx(now - 1)

    @Slot()
    def slideInNext(self):
        now = self.currentIndex()
        if self.m_wrap or now < (self.count() - 1):
            self.slideInIdx(now + 1)

    def slideInIdx(self, idx: int):
        if idx > (self.count() - 1):
            idx = idx % self.count()
        elif idx < 0:
            idx = (idx + self.count()) % self.count()
        self.slideInWgt(self.widget(idx))

    def slideInIndex(self, index: int):
        self.slideInIdx(index)

    def slideInWgt(self, widget):
        if self.m_active:
            return

        self.m_active = True

        _now = self.currentIndex()
        _next = self.indexOf(widget)

        if _now == _next:
            self.m_active = False
            return

        offsetx, offsety = (
            self.frameRect().width() + 500,
            self.frameRect().height() + 500,
        )

        self.widget(_next).setGeometry(self.frameRect())

        if not self.m_direction == Qt.Orientation.Horizontal:
            if _now < _next:
                offsetx, offsety = 0, -offsety
            else:
                offsetx = 0
        else:
            if _now < _next:
                offsetx, offsety = -offsetx, 0
            else:
                offsety = 0

        pnext = self.widget(_next).pos()
        pnow = self.widget(_now).pos()

        self.m_pnow = pnow if self.m_pnow is None else self.m_pnow

        offset = QPoint(offsetx, offsety)

        self.widget(_next).move(pnext - offset)
        self.widget(_next).show()
        self.widget(_next).raise_()

        anim_group = QParallelAnimationGroup(self, finished=self.animationDoneSlot)

        for (index, start, end) in zip(
            (_now, _next), (pnow, pnext - offset), (pnow + offset, pnext)
        ):
            animation = QPropertyAnimation(
                self.widget(index),
                b"pos",
                duration=self.m_speed,
                easingCurve=self.m_animationtype,
                startValue=start,
                endValue=end,
            )

            anim_group.addAnimation(animation)

        self.m_next = _next
        self.m_now = _now
        self.m_active = True

        anim_group.start(QAbstractAnimation.DeletionPolicy.DeleteWhenStopped)

    @Slot()
    def animationDoneSlot(self):
        self.setCurrentIndex(self.m_next)
        self.widget(self.m_now).hide()
        self.widget(self.m_now).move(self.m_pnow)
        self.m_active = False

    def setIndex(self, index: int):
        self.slideInIdx(index)

    def setWidget(self, widget: object):
        self.slideInWgt(widget)

    def setNext(self):
        self.slideInNext()

    def setPrev(self):
        self.slideInPrev()


class SlidingStackedWidget(WidgetBase, _SlidingStackedWidget):
    def __init__(self, 
                 *,
                 children: Union[list[QWidget], dict[str, QWidget]] = None,
                 direction: Qt.Orientation = Qt.Orientation.Horizontal,
                 easing_curve: QEasingCurve = QEasingCurve.Type.OutCubic,
                 speed: int = 500,
                 current: int = 0,
                 next: int = 0,
                 wrap: bool = False,
                 pnow = None,
                 active: bool = False,
                 **kwargs):
        _SlidingStackedWidget.__init__(self,
                                       None,
                                       direction,
                                       easing_curve,
                                       speed,
                                       current,
                                       next,
                                       wrap,
                                       pnow,
                                       active)
        super().__init__(**kwargs)

        if isinstance(children, dict):
            self._routes = Routes(self, "/")
            self.__setattr__("route", self._routes)
            self.__setattr__("push", lambda route: self._routes.push(route))

        if children is not None:
            self._set_children(children)

    def _set_children(self, children) -> None:
        if not isinstance(children, dict):
            for child in children:
                self.addWidget(child)
        else:
            for route, child in children.items():
                if child is not None:
                    self._routes.add(route, child)


if __name__ == "__main__":
    import sys
    import random
    from PySide6 import QtWidgets, QtGui, QtCore
    
    class MainWindow(QtWidgets.QMainWindow):
        def __init__(self, parent=None):
            super(MainWindow, self).__init__(parent)

            slidingStacked = _SlidingStackedWidget()
            for i in range(10):
                label = QtWidgets.QLabel(
                    "Qt is cool " + i * "!", alignment=QtCore.Qt.AlignCenter
                )
                color = QtGui.QColor(*random.sample(range(255), 3))
                label.setStyleSheet(
                    "QLabel{ background-color: %s; color : white; font: 40pt}"
                    % (color.name(),)
                )
                slidingStacked.addWidget(label)

            button_prev = QtWidgets.QPushButton(
                "Previous", pressed=slidingStacked.slideInPrev
            )
            button_next = QtWidgets.QPushButton(
                "Next", pressed=slidingStacked.slideInNext
            )

            hlay = QtWidgets.QHBoxLayout()
            hlay.addWidget(button_prev)
            hlay.addWidget(button_next)

            central_widget = QtWidgets.QWidget()
            self.setCentralWidget(central_widget)
            lay = QtWidgets.QVBoxLayout(central_widget)
            lay.addLayout(hlay)
            lay.addWidget(slidingStacked)

    app = QtWidgets.QApplication(sys.argv)
    w = MainWindow()
    w.resize(640, 480)
    w.show()
    sys.exit(app.exec_())



from collections import Counter
from typing import Any
from PIL import Image



class ImageColors:
    def __init__(self, image: str):

        self._image_path = image
        self._image = Image.open(image)

        # Resize the image to improve performance
        self._image = self._image.resize((100, 100))
        self._image = self._image.convert('RGB')

        # Count the occurrences of each color
        self._color_counter = Counter(self._image.getdata())

    @property
    def image_path(self) -> str:
        return self._image_path

    @property
    def total(self):
        return self._color_counter.total()

    @property
    def count(self):
        return self._color_counter.__len__()

    @property
    def colors(self) -> Any:
        return self._image.getdata()

    @property
    def most_common(self, count: int = None) -> list[tuple[tuple[int], int]]:
        '''
        count: int
            count == None, default = cls.total
        '''
        return self._color_counter.most_common(count
                                               if count is not None
                                               else self._color_counter.total())




from PySide6.QtWidgets import QGraphicsOpacityEffect, QWidget
from ..core import QObjectBase


class OpacityEffect(QObjectBase, QGraphicsOpacityEffect):
    def __init__(self, **kwargs) -> QWidget:
        QGraphicsOpacityEffect.__init__(self)
        super().__init__(**kwargs)


#!/usr/bin/python3
# coding:utf-8

from typing import Union
from PySide6.QtWidgets import QDialog, QGraphicsOpacityEffect, QGridLayout, QWidget, QApplication
from PySide6.QtCore import QEasingCurve, QPropertyAnimation, QSize, QTimer, Qt, QEvent
from PySide6.QtGui import QColor, QResizeEvent
from ...utils.theme_detect import isDark
from ...core import WidgetBase


class _MaskDialog(WidgetBase, QDialog):
    """ Dialog box base class with a mask """

    def __init__(self, 
                 parent: QWidget,
                 child: QWidget,
                 mask: QWidget = None,
                 auto_close: bool = False,
                 padding: Union[QSize, tuple, int] = 90,
                 timeout: int = None,
                 **kwargs):
        QDialog.__init__(self, parent)
        super().__init__(**kwargs)

        self._grid_layout = QGridLayout(self)
        self._auto_close = auto_close
        self._timeout = timeout
        self._padding = padding

        child.closeEvent = self._child_close_event

        if isinstance(padding, int):
            self._padding = QSize(padding, padding)

        if isinstance(padding, (tuple, list)):
            self._padding = QSize(*padding[:2])

        if child is not None:
            self.child = child

        if mask is not None:
            self._mask = mask
            self._mask.setParent(self)
        else:
            self._mask = QWidget(self)
            color = QColor(0, 0, 0, 76)
            self._mask.setStyleSheet(f"""
                background: rgba({color.red()}, \
                    {color.blue()}, \
                        {color.green()}, \
                            {color.alpha()})
                            """)

        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setGeometry(0, 0, parent.width(), parent.height())

        c = 0 if isDark() else 255
        self._mask.setStyleSheet(f'background: rgba({c}, {c}, {c}, 0.6)')
        self._mask.resize(self.size())

        self._grid_layout.addWidget(self.child)

        self.window().installEventFilter(self)

    def _child_close_event(self, e):
        self.close()

    def showEvent(self, e):
        """ fade in """
        opacityEffect = QGraphicsOpacityEffect(self)
        self.setGraphicsEffect(opacityEffect)

        opacityAni = QPropertyAnimation(opacityEffect, b'opacity', self)
        opacityAni.setStartValue(0)
        opacityAni.setEndValue(1)
        opacityAni.setDuration(200)
        opacityAni.setEasingCurve(QEasingCurve.Type.InSine)
        opacityAni.finished.connect(opacityEffect.deleteLater)
        opacityAni.start()

        super().showEvent(e)

        if self._timeout is not None:
            QTimer.singleShot(self._timeout, self.close)

    def closeEvent(self, e):
        """ fade out """
        self.child.setGraphicsEffect(None)

        opacityEffect = QGraphicsOpacityEffect(self)
        self.setGraphicsEffect(opacityEffect)

        opacityAni = QPropertyAnimation(opacityEffect, b'opacity', self)
        opacityAni.setStartValue(1)
        opacityAni.setEndValue(0)
        opacityAni.setDuration(100)
        opacityAni.setEasingCurve(QEasingCurve.Type.OutCubic)
        opacityAni.finished.connect(self.deleteLater)
        opacityAni.start()

        super().closeEvent(e)

    def resizeEvent(self, e):
        self.child.blockSignals(True)
        self.child.setFixedSize(self.width() -  self._padding.width(),
                                self.height() - self._padding.height())
        self.child.blockSignals(False)
        self._mask.resize(self.size())

    def eventFilter(self, obj, e: QEvent):
        if obj is self.window():
            if e.type() == QEvent.Type.Resize:
                re = QResizeEvent(e)
                self.resize(re.size())

            if e.type() == QEvent.Type.MouseButtonPress and self._auto_close:
                self.close()

        return super().eventFilter(obj, e)


class MaskDialog:
    @classmethod
    def display(cls,
                *,
                child: QWidget,
                parent: QWidget = None,
                mask: QWidget = None,
                auto_close: bool = False,
                padding: Union[QSize, tuple, int] = 90,
                timeout: int = None,
                **kwargs):

        if not parent:
            parent = QApplication.activeWindow()

        mask_dialog = _MaskDialog(parent, 
                                  child, 
                                  mask, 
                                  auto_close,
                                  padding,
                                  timeout,
                                  **kwargs)
        mask_dialog.show()


from PySide6.QtWidgets import QGraphicsDropShadowEffect
from ..core import QObjectBase


class DropShadowEffect(QObjectBase, QGraphicsDropShadowEffect):
    def __init__(self, **kwargs):
        QGraphicsDropShadowEffect.__init__(self)
        super().__init__(**kwargs)


from PySide6.QtWidgets import QWidget, QLayout, QLayoutItem
from PySide6.QtCore import Qt
from typing import Union


class GridAlignment:
    def __init__(self, 
                 child: QWidget,
                 alignment: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignCenter) -> None:

        self._child = child
        self._alignment = alignment


class GridLayoutItemWrapper:
    def __init__(
        self,
        *,
        child: Union[QWidget, QLayout, QLayoutItem],
        row: int = None,
        column: int = None,
        row_span: int = None,
        column_span: int = None,
        alignment: Qt.AlignmentFlag = None
    ) -> None:

        if not child:
            raise AttributeError("invalid child widget!")

        self._attrs = {}

        for k, v in zip(("arg__1", "row", "column", "rowSpan", "columnSpan", "alignment"), 
                        (child, row, column, row_span, column_span, alignment)):
            if v is not None:
                self._attrs[k] = v

    @property
    def child(self) -> QWidget:
        return self.attrs.get("arg__1")

    @property
    def attrs(self) -> dict:
        return self._attrs


#!/usr/bin/python3
# -*- coding: utf-8 -*-

from PySide6.QtCore import QRect, Qt, QRectF, QLineF
from PySide6.QtGui import QColor, QPainter, QPen, QTransform, QPainterPath
from PySide6.QtWidgets import QProgressBar, QStyleOptionProgressBar
from ..animation.style_animation import ProgressStyleAnimation
from ..core import WidgetBase


class _ColourfulProgressBar(QProgressBar):
    def __init__(self, 
                 color: QColor = QColor(43, 194, 83),
                 fps: int = 60,
                 line_width: int = 50,
                 radius: int = None):
        super(_ColourfulProgressBar, self).__init__()

        self._color = color
        self._fps = fps
        self._lineWidth = line_width
        self._radius = radius
        self._animation = None

    def setColor(self, color: QColor):
        self._color = color
        self.update()

    def setFps(self, fps: int):
        self._fps = max(int(fps), 1)
        self.update()

    def setLineWidth(self, width: int):
        self._lineWidth = max(int(width), 0)
        self.update()

    def setRadius(self, radius: int):
        self._radius = max(int(radius), 1)
        self.update()

    def paintEvent(self, _):
        option = QStyleOptionProgressBar()
        self.initStyleOption(option)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.translate(0.5, 0.5)

        rect = self.rect()
        vertical = self.orientation() == Qt.Orientation.Vertical
        inverted = self.invertedAppearance()
        indeterminate = ((self.minimum() == self.maximum()) 
                         or (self.minimum() < self.value() < self.maximum()))

        if vertical:
            rect = QRect(rect.left(), rect.top(), rect.height(),rect.width())
            m = QTransform.fromTranslate(rect.height(), 0)
            m.rotate(90.0)
            painter.setTransform(m, True)

        maxWidth = rect.width()
        progress = max(self.value(), self.minimum())
        totalSteps = max(1, self.maximum() - self.minimum())
        progressSteps = progress - self.minimum()
        progressBarWidth = int(progressSteps * maxWidth / totalSteps)
        width = progressBarWidth
        radius = max(1, (min(width,
                             self.width() if vertical else self.height()) //
                         4) if self._radius is None else self._radius)

        reverse = (not vertical and
                   self.layoutDirection() == Qt.LayoutDirection.RightToLeft) or vertical
        if inverted:
            reverse = not reverse

        path = QPainterPath()
        if not reverse:
            progressBar = QRectF(rect.left(), rect.top(), width, rect.height())
        else:
            progressBar = QRectF(rect.right() - width, rect.top(), width,
                                 rect.height())

        path.addRoundedRect(progressBar, radius, radius)
        painter.setClipPath(path)

        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self._color)
        painter.drawRoundedRect(progressBar, radius, radius)

        if not indeterminate:
            if self._animation:
                self._animation.stop()
                self._animation = None
        else:
            color = self._color.lighter(320)
            color.setAlpha(80)
            painter.setPen(QPen(color, self._lineWidth))

            if self._animation:
                step = int(self._animation.animationStep() % self._lineWidth)
            else:
                step = 0
                self._animation = ProgressStyleAnimation(self._fps, self)
                self._animation.start()

            startX = int(progressBar.left() - rect.height() - self._lineWidth)
            endX = int(rect.right() + self._lineWidth)

            if (not inverted and not vertical) or (inverted and vertical):
                lines = [
                    QLineF(x + step, progressBar.bottom(),
                           x + rect.height() + step, progressBar.top())
                    for x in range(startX, endX, self._lineWidth)
                ]
            else:
                lines = [
                    QLineF(x - step, progressBar.bottom(),
                           x + rect.height() - step, progressBar.top())
                    for x in range(startX, endX, self._lineWidth)
                ]
            painter.drawLines(lines)


class ColourfulProgressBar(WidgetBase, _ColourfulProgressBar):
    def __init__(self, 
                 *,
                 color: QColor = QColor(43, 194, 83),
                 fps: int = 60,
                 line_width: int = 50,
                 radius: int = None,
                 **kwargs):
        _ColourfulProgressBar.__init__(self, color, fps, line_width, radius)
        super().__init__(**kwargs)


from enum import Enum, auto
from typing import Callable, Optional
from PySide6.QtCore import QRect, QSize, QTimer, Qt
from PySide6.QtGui import QColor, QIcon, QMouseEvent, QPainter, QPaintEvent
from PySide6.QtWidgets import QGridLayout, QLabel, QToolButton, QWidget
from ..teaching_tip.tool_tip import _TeachingTip
from ...enums.teaching_tip_tails import TeachingTipTailPositions


class SideBarButton(QToolButton):
    class Alignment(Enum):
        top = auto()
        bottom = auto()

    def __init__(
        self, 
        icon: QIcon,
        tool_tip: Optional[str] = None,
        radius: int = 8,
        parent: Optional[QWidget] = None,
        bg_color: QColor = QColor("#191E23"),
        on_color: QColor = QColor("#282D32"),
        edge_color: QColor = QColor("#0066FF"),
        icon_off: QIcon = None,
        icon_on: QIcon = None,
        clicked: Callable = None
    ) -> None:
        super().__init__(parent)

        if clicked is not None:
            self.clicked.connect(clicked)

        # Set attributes from arguments
        self._icon = icon
        self._radius = radius
        self._tool_tip: Optional[_TeachingTip] = None

        self._icon_off = icon_off if icon_off is not None else icon
        self._icon_on = icon_on if icon_on is not None else icon

        if tool_tip:
            _tool_tip_widget = QWidget()
            _tool_tip_grid = QGridLayout(_tool_tip_widget)
            _label = QLabel(tool_tip)
            _label.setWordWrap(True)
            _tool_tip_grid.addWidget(_label)
            _tool_tip_widget.setLayout(_tool_tip_grid)

            self._tool_tip = _TeachingTip(
                _tool_tip_widget,
                self,
                TeachingTipTailPositions.left
            )

        self.setBackgroundColor(bg_color)
        self.setOnColor(on_color)
        self.setEdgeColor(edge_color)

        # Set QToolButton properties
        self.setCheckable(True)
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setMouseTracking(True)

        self._timer = QTimer(self)
        self._timer.timeout.connect(self._start_tooltip)

    def _start_tooltip(self):
        if self._tool_tip.isHidden():
            self._tool_tip.show()

    def enterEvent(self, arg__1) -> None:
        if self._tool_tip and self.underMouse():
            self._timer.start(1000)

        return super().enterEvent(arg__1)

    def leaveEvent(self, arg__1) -> None:
        if self._tool_tip:
            self._timer.stop()
            self._tool_tip.hide()

        return super().leaveEvent(arg__1)

    def setBackgroundColor(self, color: QColor = QColor("#191E23")):
        self._bg_color = color
        self.update()

    def setEdgeColor(self, color: QColor = QColor("#0066FF")):
        self._h1_color = color
        self.update()

    def setOnColor(self, color: QColor = QColor("#282D32")):
        self._on_color = color
        self.update()

    def minimumSizeHint(self) -> QSize:
        return QSize(32, 32)

    def mousePressEvent(self, arg__1: QMouseEvent) -> None:
        if self._tool_tip:
            self._timer.stop()
            self._tool_tip.hide()
        return super().mousePressEvent(arg__1)

    def paintEvent(self, event: QPaintEvent) -> None:
        super().paintEvent(event)

        # Define areas for painting
        radius = self._radius
        width = self.width()
        height = self.height()

        self.rect_background = QRect(0, 0, width, height)
        self.rect_icon = QRect(width // 4, height // 4, width // 2, height // 2)
        self.rect_active_right = QRect(width - 2 * radius, 0, 2 * radius, height)

        self.rect_active_left = QRect(radius // 2,
                                      radius, 
                                      2 * radius, 
                                      height - 2 * radius)

        self.rect_active_middle = QRect(radius, 
                                        radius, 
                                        width - 2 * radius, 
                                        height - 2 * radius)


        self.rect_active_right_corner_1 = QRect(width - 3 * radius,
                                                -radius,
                                                3 * radius,
                                                2 * radius)

        self.rect_active_right_corner_2 = QRect(width - 3 * radius, 
                                                height - radius, 
                                                3 * radius, 
                                                2 * radius)

        painter = QPainter()
        painter.begin(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.setPen(Qt.PenStyle.NoPen)

        # Draw background
        painter.setBrush(self._bg_color)
        painter.drawRect(self.rect_background)

        if self.isChecked():
            # Draw left highlight area
            painter.setBrush(self._h1_color)
            painter.drawRoundedRect(self.rect_active_left, self._radius, self._radius)

            # Draw middle and right active area
            painter.setBrush(self._on_color)
            painter.drawRoundedRect(self.rect_active_middle, self._radius, self._radius)

            painter.setBrush(self._on_color)
            painter.drawRect(self.rect_active_right)

            # Draw right corners area (background)
            painter.setBrush(self._bg_color)
            painter.drawRoundedRect(self.rect_active_right_corner_1, 
                                    self._radius, self._radius)
            painter.drawRoundedRect(self.rect_active_right_corner_2, 
                                    self._radius, self._radius)

            # Draw icon
            self._icon_on.paint(painter, self.rect_icon)

        elif self.underMouse():
            # Draw middle active area
            painter.setBrush(self._on_color)
            painter.drawRoundedRect(self.rect_active_middle, self._radius, self._radius)

            # Draw icon
            self._icon_on.paint(painter, self.rect_icon)

        else:
            # Draw icon
            self._icon_off.paint(painter, self.rect_icon)

        painter.end()


#-----------------------------------------------------------------------------
#  Copyright (C) 2019 Alberto Sottile, Eric Larson
#
#  Distributed under the terms of the 3-clause BSD License.
#-----------------------------------------------------------------------------

import subprocess
from ...enums.theme import Theme


def theme() -> Theme:
    # Here we just triage to GTK settings for now
    try:
        out = subprocess.run(
            ['gsettings', 'get', 'org.gnome.desktop.interface', 'gtk-theme'],
            capture_output=True)
        stdout = out.stdout.decode()
    except Exception:
        return Theme.light
    # we have a string, now remove start and end quote
    theme = stdout.lower().strip()[1:-1]
    if '-dark' in theme.lower():
        return Theme.dark
    else:
        return Theme.light

def isDark():
    return theme() == Theme.dark

def isLight():
    return theme() == Theme.light

# def listener(callback: typing.Callable[[enums.Theme], None]) -> None:
def listener(callback):
    with subprocess.Popen(
        ('gsettings', 'monitor', 'org.gnome.desktop.interface', 'gtk-theme'),
        stdout=subprocess.PIPE,
        universal_newlines=True,
    ) as p:
        for line in p.stdout:
            callback(Theme.dark 
                     if '-dark' in line.strip().removeprefix("gtk-theme: '").removesuffix("'").lower() 
                     else Theme.light)


import os.path
from imp import load_package


class Imports:
    @staticmethod
    def import_(path: str) -> object:
        """python function for import module from files"""
        return load_package(os.path.basename(path).split(".", 1)[0], path)

    @staticmethod
    def include(path: str) -> object:
        """python function for import module from files"""
        return Imports.import_(path)

    @staticmethod
    def import_plugin(path: str) -> object:
        """python function for import module from plugins"""
        return load_package(os.path.basename(path).split(".")[-1], path)

    @staticmethod
    def method(path: str, 
               name: str, 
               default: object = None) -> object:
        """
        >>> py = method("/path/to/script.py", "method")
        >>> py() # call method
        ...
        """
        return getattr(Imports.import_(path), name, default)


#!/usr/bin/python3

from enum import Enum
from PySide6.QtGui import QIcon, QIconEngine, QImage, QPixmap


class AbstractIcon(Enum):
    '''
    class Icons(AbstractIcon):
        icon_name: Union[str, QIcon, QIconEngine, QPixmap, QImage]
    '''

    def __new__(cls, value):
        obj = object.__new__(cls)
        if not cls._check_type(value):
            raise ValueError("invalid enum value, " \
                "supported value str | QIcon | QIconEngine | QPixmap | QImage")
        obj._value_ = value
        return obj

    @classmethod
    def _check_type(cls, value):
        for _type in (
            str,
            QIcon,
            QIconEngine,
            QPixmap,
            QImage
        ):
            if (_check := ((type(value) == _type) 
                           or isinstance(value, _type))):
                return _check
        return _check

    @property
    def icon(self) -> QIcon:
        return QIcon(self.value)

    @property
    def pixmap(self) -> QPixmap:
        if not isinstance(self.value, (QPixmap, QImage, str)):
            raise ValueError("invalid value, " \
                "supported value QPixmap | QImage | str")
        return QPixmap(self.value)

    @property
    def image(self) -> QImage:
        if not isinstance(self.value, (QImage, str)):
            raise ValueError("invalid value, \
                supported value QImage | str")
        return QPixmap(self.value)


class IconBase(AbstractIcon):
    pass


from typing import Any, Union
from PySide6.QtCore import Qt, QObject
from PySide6.QtWidgets import QApplication, QWidget
from ..enums.widgets import Widgets, WidgetTypeVar
from .base import (
    BehaviorDeclarative,
    TrackingDeclarative
)


class Api:
    '''

    ```
    api = Api(parent)
    obj = api.get('uid')
    >> obj = object
    ```

    ### Direct Access
    ```
    Api.fetch('uid')
    >> object
    ```
    '''
    def __init__(self, 
                 parent: QObject = None,
                 w_type: WidgetTypeVar | object = Widgets.any):

        self.parent = parent
        self.w_type = w_type

    @classmethod
    def _find_child(cls, 
                    target: QObject,
                    uid: str,
                    w_type: WidgetTypeVar) -> QObject | None:

        return target.findChild(cls._get_wtype(w_type 
                                                if w_type is not None
                                                else cls.w_type),
                                uid.strip(), 
                                Qt.FindChildOption.FindChildrenRecursively)

    def _get_wtype(self, w_type: WidgetTypeVar = Widgets.any):
        return (w_type.value 
                if isinstance(w_type, Widgets) 
                else w_type)

    def get(self, 
            uid: str, 
            _type: WidgetTypeVar | object = Widgets.any) -> QObject | None:

        if Api.isinstance(_type, BehaviorDeclarative):
            return TrackingDeclarative.get(uid)

        if self.parent is not None:
            return self._find_child(self.parent, uid, _type)

        for parent in QApplication.topLevelWidgets():
            widget = self._find_child(parent, uid, _type)
            if widget is not None:
                return widget

    @staticmethod
    def isinstance(obj, _type) -> bool:
        if obj == _type:
            return True

        if type(obj) == _type:
            return True

        if isinstance(obj, _type):
            return True

        return False

    @classmethod
    def fetch(cls,
              uid: str,
              _type: WidgetTypeVar | object = Widgets.any
        ) -> Union[QWidget, QObject, Any, None]:

        uid = uid.strip()

        if uid == QApplication.instance().objectName().strip():
            return QApplication.instance()

        if cls.isinstance(_type, BehaviorDeclarative):
            return TrackingDeclarative.get(uid)

        if uid in TrackingDeclarative._instances_dict:
            return TrackingDeclarative.get(uid)

        for parent in QApplication.topLevelWidgets():
            if uid == parent.objectName().strip():
                return parent

            if (widget := Api._find_child(parent, uid, _type)) is not None:
                return widget

    @staticmethod
    def declarative_fetch(uid: str) -> Any:
        return TrackingDeclarative.get(uid)

    @staticmethod
    def dec_fetch(uid: str) -> Any:
        return Api.declarative_fetch(uid)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return self.get(*args, **kwds)


from typing import Optional
from PySide6.QtGui import QColor, QResizeEvent
from PySide6.QtWidgets import QFrame, QSizePolicy, QSpacerItem, QHBoxLayout, QWidget
from .nav_bar_button import NavBarButton


_BG_FRAME_STYLESHEET = """\
QFrame#_nav_bar_bg_frame {{
    background-color: {bg_color};
    border: none;
    border-radius: 8px;
}}
"""


class NavBar(QWidget):
    def __init__(self, 
                 parent: Optional[QWidget] = None,
                 bg_color: QColor = QColor("#191E23")) -> None:
        super().__init__(parent)

        # Define attributes
        self._first_button = True
        self._left_buttons: list[NavBarButton] = []
        self._right_buttons: list[NavBarButton] = []
        self._center_buttons: list[NavBarButton] = []

        # Layout for this widget
        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        # Background frame and its layout
        self._bg_frame = QFrame(self)
        self._bg_frame.setObjectName("_nav_bar_bg_frame")

        self._bg_layout = QHBoxLayout(self._bg_frame)
        self._bg_layout.setContentsMargins(15, 0, 15, 0)
        self._bg_layout.setSpacing(0)

        self._layout.addWidget(self._bg_frame)

        # Left nav bar buttons frame and its layout
        self._left_frame = QFrame(self._bg_frame)
        self._left_frame_layout = QHBoxLayout(self._left_frame)
        self._left_frame_layout.setContentsMargins(0, 0, 0, 0)
        self._left_frame_layout.setSpacing(0)
        self._bg_layout.addWidget(self._left_frame)

        # Horizontal spacer between left and right bar
        self._horizontal_spacer = QSpacerItem(0, 0, 
                                            QSizePolicy.Policy.Expanding,
                                            QSizePolicy.Policy.Minimum)
        self._bg_layout.addItem(self._horizontal_spacer)

        # Left nav bar buttons frame and its layout
        self._center_frame = QFrame(self._bg_frame)
        self._center_frame_layout = QHBoxLayout(self._center_frame)
        self._center_frame_layout.setContentsMargins(0, 0, 0, 0)
        self._center_frame_layout.setSpacing(0)
        self._bg_layout.addWidget(self._center_frame)

        # Horizontal spacer between left and right bar
        self._horizontal_spacer = QSpacerItem(0, 0, 
                                            QSizePolicy.Policy.Expanding,
                                            QSizePolicy.Policy.Minimum)
        self._bg_layout.addItem(self._horizontal_spacer)

        # Bottom side bar buttons frame and its layout
        self._right_frame = QFrame(self._bg_frame)
        self._right_frame_layout = QHBoxLayout(self._right_frame)
        self._right_frame_layout.setContentsMargins(0, 0, 0, 0)
        self._right_frame_layout.setSpacing(0)
        self._bg_layout.addWidget(self._right_frame)

        self.setBackgroundColor(bg_color)

    def setBackgroundColor(self, color: QColor("#191E23")) -> None:
        self._bg_frame.setStyleSheet(_BG_FRAME_STYLESHEET.format(bg_color=color.name()))

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)

        buttons: list[NavBarButton] = self.findChildren(NavBarButton)
        for button in buttons:
            button.setFixedSize(self.height(), self.height())

    def addButton(self,
                  button: NavBarButton, 
                  alignment: NavBarButton.Alignment = NavBarButton.Alignment.left) -> None:

        button.setParent(self)
        button.clicked.connect(self._buttonCallback)  # type: ignore[attr-defined]

        if alignment == NavBarButton.Alignment.left:
            self._left_buttons.append(button)
            self._left_frame_layout.addWidget(button)

        elif alignment == NavBarButton.Alignment.center:
            self._center_buttons.append(button)
            self._center_frame_layout.addWidget(button)

        elif alignment == NavBarButton.Alignment.right:
            self._right_buttons.append(button)
            self._right_frame_layout.addWidget(button)

        else:
            raise ValueError(
                "Invalid value for 'alignment'!"
                "Supported values are 'NavBarButton.Alignment.left' and 'NavBarButton.Alignment.right'."
            )

        # Select the first button as default
        if self._first_button:
            button.click()
            self._first_button = False

    def _buttonCallback(self):
        buttons: list[NavBarButton] = self.findChildren(NavBarButton)
        for button in buttons:
            if button == self.sender():
                button.setChecked(True)
            else:
                button.setChecked(False)


from PySide6.QtQuick import QQuickView
from PySide6.QtCore import QUrl
from ...core import QObjectBase


class QuickView(QObjectBase, QQuickView):
    def __init__(self,
                 *,
                 qml: QUrl | str,
                 **kwargs):
        QQuickView.__init__(self, qml)
        super().__init__(**kwargs)

        self.setResizeMode(QQuickView.ResizeMode.SizeRootObjectToView)


#!/usr/bin/python
# -*- coding: utf-8 -*-

from PySide6.QtCore import (
    QSequentialAnimationGroup, 
    QPauseAnimation, 
    QPropertyAnimation,
    QParallelAnimationGroup, 
    QObject, 
    QSize, 
    Qt, 
    QRectF, 
    Signal, 
    Property
)
from PySide6.QtGui import QPainter, QColor
from PySide6.QtWidgets import QWidget
from ..core import WidgetBase


def _qBound(miv, cv, mxv):
    return max(min(cv, mxv), miv)


class CircleItem(QObject):
    _x: int = 0 
    _opacity: int = 1
    value_changed = Signal()

    @Property(float)
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, x: float):
        self._x = x
        self.value_changed.emit()

    @Property(float)
    def opacity(self) -> float:
        return self._opacity

    @opacity.setter
    def opacity(self, opacity: float):
        self._opacity = opacity


class _MetroCircleProgress(QWidget):
    radius: int = 5
    color = QColor(24, 189, 155)
    bg_color = QColor(Qt.GlobalColor.transparent)

    def __init__(self, 
                 radius: int = 5, 
                 color: QColor = QColor(24, 189, 155),
                 bg_color: QColor = QColor(Qt.GlobalColor.transparent)):
        super(_MetroCircleProgress, self).__init__()

        self.radius = radius
        self.color = color
        self.bg_color = bg_color
        self._items = []

        self._initAnimations()

    def setRadius(self, radius: int):
        if self.radius != radius:
            self.radius = radius

    def setColor(self, color: QColor):
        if self.color != color:
            self.color = color
            self.update()

    def setBackgroundColor(self, bg_color: QColor):
        if self.bg_color != bg_color:
            self.bg_color = bg_color
            self.update()

    def paintEvent(self, event):
        super(_MetroCircleProgress, self).paintEvent(event)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.fillRect(self.rect(), self.bg_color)
        painter.setPen(Qt.PenStyle.NoPen)

        for item, _ in self._items:
            painter.save()
            color = self.color.toRgb()
            color.setAlphaF(item.opacity)
            painter.setBrush(color)
            # 5<= radius <=10
            radius = _qBound(self.radius, self.radius / 200 *
                            self.height(), 2 * self.radius)
            diameter = 2 * radius
            painter.drawRoundedRect(
                QRectF(
                    item.x / 100 * self.width() - diameter,
                    (self.height() - radius) / 2,
                    diameter, diameter
                ), radius, radius)
            painter.restore()

    def _initAnimations(self):
        for index in range(5):
            item = CircleItem(self)
            item.value_changed.connect(self.update)

            seqAnimation = QSequentialAnimationGroup(self)
            seqAnimation.setLoopCount(-1)
            self._items.append((item, seqAnimation))

            seqAnimation.addAnimation(QPauseAnimation(150 * index, self))

            parAnimation1 = QParallelAnimationGroup(self)
            parAnimation1.addAnimation(QPropertyAnimation(
                item, b'opacity', self, duration=400, startValue=0, endValue=1.0))
            parAnimation1.addAnimation(QPropertyAnimation(
                item, b'x', self, duration=400, startValue=0, endValue=25.0))
            seqAnimation.addAnimation(parAnimation1)

            seqAnimation.addAnimation(QPropertyAnimation(
                item, b'x', self, duration=2000, startValue=25.0, endValue=75.0))

            parAnimation2 = QParallelAnimationGroup(self)
            parAnimation2.addAnimation(QPropertyAnimation(
                item, b'opacity', self, duration=400, startValue=1.0, endValue=0))
            parAnimation2.addAnimation(QPropertyAnimation(
                item, b'x', self, duration=400, startValue=75.0, endValue=100.0))
            seqAnimation.addAnimation(parAnimation2)

            seqAnimation.addAnimation(
                QPauseAnimation((5 - index - 1) * 150, self))

        for _, animation in self._items:
            animation.start()

    def sizeHint(self):
        return QSize(100, self.radius * 2)


class MetroCircleProgress(WidgetBase, _MetroCircleProgress):
    def __init__(self,
                 *,
                 radius: int = 5, 
                 color: QColor = QColor(24, 189, 155),
                 bg_color: QColor = QColor(Qt.GlobalColor.transparent), 
                 **kwargs):
        _MetroCircleProgress.__init__(self, radius, color, bg_color)
        super().__init__(**kwargs)


from PySide6.QtWidgets import QToolButton
from PySide6.QtCore import Signal
from ..core import WidgetBase


class ToolButton(WidgetBase, QToolButton):
    long_press = Signal()
    long_repeat = Signal()
    long_release = Signal()
    long_click = Signal()

    def __init__(self, 
                 *,
                 enable_long_press: bool = False,
                 **kwargs):
        QToolButton.__init__(self)
        super().__init__(**kwargs)

        self._state = 0

        if enable_long_press:
            self.setAutoRepeat(True)
            self.setAutoRepeatDelay(1000)
            self.setAutoRepeatInterval(200)

        self.clicked.connect(self._handleLongClicked)

    def _handleLongClicked(self):
        if self.isDown():
            if self._state == 0:
                self._state = 1
                self.setAutoRepeatInterval(50)
                self.long_press.emit()
            else:
                self.long_repeat.emit()
        elif self._state == 1:
            self._state = 0
            self.setAutoRepeatInterval(200)
            self.long_release.emit()
        else:
            self.long_click.emit()